"""AsyncFIX errors module."""


class FIXError(Exception):
    """Generic AsyncFIX error."""


class FIXMessageError(FIXError):
    """FIXMessage related error."""


class FIXConnectionError(FIXError):
    """FIX connection or session error."""


class DuplicateSeqNoError(FIXError):
    """Journaler duplicated seq no written (critical error)."""


class EncodingError(FIXError):
    """Codec encoding/decoding error."""


class TagNotFoundError(FIXMessageError):
    """Requested Tag not present in message."""


class DuplicatedTagError(FIXMessageError):
    """Trying to set tag which is already exist."""


class RepeatingTagError(FIXMessageError):
    """Tag was repeated after decoding, indicates mishandled fix group."""


class UnmappedRepeatedGrpError(FIXMessageError):
    """Repeating group improperly set up by protocol."""



import asyncio
import logging
import os
import time
import warnings
import xml.etree.ElementTree as ET
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
import pytest_asyncio

from asyncfix import FIXMessage, FIXTester, FMsg, FTag
from asyncfix.connection import AsyncFIXConnection, ConnectionRole, ConnectionState
from asyncfix.errors import FIXConnectionError
from asyncfix.journaler import Journaler
from asyncfix.message import MessageDirection
from asyncfix.protocol import FIXProtocol44, FIXSchema
from asyncfix.protocol.order_single import FIXNewOrderSingle

TEST_DIR = os.path.abspath(os.path.dirname(__file__))
fix44_schema = ET.parse(os.path.join(TEST_DIR, "FIX44.xml"))
FIX_SCHEMA = FIXSchema(fix44_schema)


@pytest_asyncio.fixture
async def fix_msg():
    msg = FIXMessage(FMsg.NEWORDERSINGLE)
    msg.set(FTag.Price, "123.45")
    msg.set(FTag.OrderQty, 9876)
    msg.set(FTag.Symbol, "VOD.L")
    return msg


@pytest_asyncio.fixture
async def fix_connection_socket():
    log = logging.getLogger("asyncfix_test")
    log.setLevel(logging.DEBUG)
    j = Journaler()
    connection = AsyncFIXConnection(
        FIXProtocol44(),
        "INITIATOR",
        "ACCEPTOR",
        journaler=j,
        host="localhost",
        port="64444",
        heartbeat_period=30,
        logger=log,
    )
    connection._connection_state = ConnectionState.NETWORK_CONN_ESTABLISHED
    assert connection._connection_role == ConnectionRole.UNKNOWN
    connection._socket_reader = MagicMock()
    connection._socket_reader.read = AsyncMock()

    connection._socket_writer = MagicMock()
    connection._socket_writer.close = MagicMock()
    connection._socket_writer.wait_closed = AsyncMock()

    return connection


@pytest_asyncio.fixture
async def fix_connection():
    log = logging.getLogger("asyncfix_test")
    log.setLevel(logging.DEBUG)
    j = Journaler()
    connection = AsyncFIXConnection(
        FIXProtocol44(),
        "INITIATOR",
        "ACCEPTOR",
        journaler=j,
        host="localhost",
        port="64444",
        heartbeat_period=30,
        logger=log,
    )
    connection._connection_state = ConnectionState.NETWORK_CONN_ESTABLISHED
    assert connection._connection_role == ConnectionRole.UNKNOWN
    return connection


@pytest.mark.asyncio
async def test_connection_send_not_connected_error(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)
    msg = ft.msg_logon()

    for state in [
        ConnectionState.DISCONNECTED_NOCONN_TODAY,
        ConnectionState.DISCONNECTED_BROKEN_CONN,
        ConnectionState.DISCONNECTED_WCONN_TODAY,
        ConnectionState.AWAITING_CONNECTION,
        ConnectionState.INITIATE_CONNECTION,
    ]:
        conn._connection_state = state
        with pytest.raises(
            FIXConnectionError,
            match="Connection must be established before sending any FIX message",
        ):
            await conn.send_msg(msg)


@pytest.mark.asyncio
async def test_connection_send_first_logon_sets_state(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    conn._connection_state = ConnectionState.NETWORK_CONN_ESTABLISHED
    msg = ft.msg_logon()
    await conn.send_msg(msg)
    assert conn._connection_state == ConnectionState.LOGON_INITIAL_SENT

    # This is not allowed until logon confirmation!
    with pytest.raises(
        FIXConnectionError,
        match=(
            r"Initiator is waiting for Logon\(\) response, you must not send any"
            r" additional messages before"
        ),
    ):
        await conn.send_msg(ft.msg_heartbeat())


@pytest.mark.asyncio
async def test_connection_send_first_must_be_logon(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    conn._connection_state = ConnectionState.NETWORK_CONN_ESTABLISHED
    msg = ft.msg_sequence_reset(1, 12)
    with pytest.raises(
        FIXConnectionError,
        match=(
            r"You must send first Logon\(35=A\)/Logout\(\) message immediately after"
            r" connection.*"
        ),
    ):
        await conn.send_msg(msg)

    assert conn.connection_state == ConnectionState.NETWORK_CONN_ESTABLISHED


@pytest.mark.asyncio
async def test_connection_logon_acceptor_logon(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    assert conn._connection_state == ConnectionState.NETWORK_CONN_ESTABLISHED
    rmsg = await ft.reply(ft.msg_logon())
    assert conn.connection_role == ConnectionRole.ACCEPTOR
    assert conn._connection_state == ConnectionState.ACTIVE


@pytest.mark.asyncio
async def test_connection_logon_acceptor_logon_first_message_expected(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    assert conn._connection_state == ConnectionState.NETWORK_CONN_ESTABLISHED
    msg_reset = ft.msg_sequence_reset(1, 2)
    rmsg = await ft.reply(msg_reset)
    assert conn._connection_state == ConnectionState.DISCONNECTED_BROKEN_CONN


@pytest.mark.asyncio
async def test_connection_logon_valid(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    assert conn._connection_role == ConnectionRole.INITIATOR
    assert conn._connection_state == ConnectionState.LOGON_INITIAL_SENT

    assert len(ft.initiator_sent) == 1

    assert ft.initiator_sent_query((FTag.SenderCompID, FTag.TargetCompID)) == {
        FTag.SenderCompID: "INITIATOR",
        FTag.TargetCompID: "ACCEPTOR",
    }

    assert ft.initiator_sent_query((35, 34)) == {FTag.MsgType: FMsg.LOGON, "34": "1"}

    ft.reset_messages()
    assert not ft.acceptor_sent
    assert not ft.initiator_sent
    assert not ft.acceptor_rcv_que

    await ft.reply(ft.msg_logon())
    assert conn._connection_state == ConnectionState.ACTIVE
    # FIX Tester.reply() - simulated server response (SenderCompID/TargetCompID swapped)
    assert ft.acceptor_sent_query((FTag.SenderCompID, FTag.TargetCompID)) == {
        FTag.TargetCompID: "INITIATOR",
        FTag.SenderCompID: "ACCEPTOR",
    }


@pytest.mark.asyncio
async def test_connection_logon_low_seq_num_by_initator(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    conn._session.next_num_out = 20
    ft.set_next_num(num_in=21)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    assert conn._connection_role == ConnectionRole.INITIATOR
    assert conn._connection_state == ConnectionState.LOGON_INITIAL_SENT

    assert len(ft.initiator_sent) == 1

    assert ft.initiator_sent_query((FTag.SenderCompID, FTag.TargetCompID)) == {
        FTag.SenderCompID: "INITIATOR",
        FTag.TargetCompID: "ACCEPTOR",
    }

    assert ft.initiator_sent_query((35, 34)) == {FTag.MsgType: FMsg.LOGON, "34": "20"}

    await ft.process_msg_acceptor()

    assert ft.conn_accept._connection_state == ConnectionState.DISCONNECTED_BROKEN_CONN
    assert conn._connection_state == ConnectionState.DISCONNECTED_BROKEN_CONN

    assert ft.acceptor_sent_query((35, 58)) == {
        FTag.MsgType: FMsg.LOGOUT,
        "58": "MsgSeqNum is too low, expected 21, got 20",
    }


@pytest.mark.asyncio
async def test_connection_logon_low_seq_num_by_acceptor(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    assert conn._connection_role == ConnectionRole.INITIATOR
    assert conn._connection_state == ConnectionState.LOGON_INITIAL_SENT

    assert len(ft.initiator_sent) == 1

    assert ft.initiator_sent_query((FTag.SenderCompID, FTag.TargetCompID)) == {
        FTag.SenderCompID: "INITIATOR",
        FTag.TargetCompID: "ACCEPTOR",
    }

    assert ft.initiator_sent_query((35, 34)) == {FTag.MsgType: FMsg.LOGON, "34": "1"}

    conn._session.next_num_in = 10
    ft.set_next_num(num_out=4)
    await ft.process_msg_acceptor()
    assert len(ft.initiator_sent) == 2
    assert len(ft.acceptor_sent) == 1
    assert ft.acceptor_sent_query((35, 34), 0) == {FTag.MsgType: FMsg.LOGON, "34": "4"}
    assert ft.initiator_sent_query((35, 58)) == {
        FTag.MsgType: FMsg.LOGOUT,
        "58": "MsgSeqNum is too low, expected 10, got 4",
    }

    assert ft.conn_accept._connection_state == ConnectionState.DISCONNECTED_WCONN_TODAY
    assert conn._connection_state == ConnectionState.DISCONNECTED_BROKEN_CONN


@pytest.mark.asyncio
async def test_connection_validation_missing_seqnum(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    msg_out = ft.initiator_sent[-1]

    del msg_out[34]
    assert ft.conn_accept._validate_integrity(msg_out) == "MsgSeqNum(34) tag is missing"


@pytest.mark.asyncio
async def test_connection_validation_seqnum_toolow(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    conn._session.next_num_out = 20

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    msg_out = ft.initiator_sent[-1]
    ft.set_next_num(num_in=21)

    assert (
        ft.conn_accept._validate_integrity(msg_out)
        == "MsgSeqNum is too low, expected 21, got 20"
    )


@pytest.mark.asyncio
async def test_connection_validation_seqnum_toolow__possdup__in_active(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=None, connection=conn)

    conn._session.next_num_out = 20
    conn._connection_state = ConnectionState.ACTIVE

    msg = FIXMessage(FMsg.NEWS, {FTag.PossDupFlag: "Y", FTag.MsgSeqNum: 19})
    await conn.send_msg(msg)
    msg_out = ft.initiator_sent[-1]
    ft.set_next_num(num_in=21)

    assert conn._connection_state == ConnectionState.ACTIVE
    assert msg_out[FTag.PossDupFlag] == "Y"
    assert (
        ft.conn_accept._validate_integrity(msg_out)
        == "MsgSeqNum is too low, expected 21, got 19"
    )


@pytest.mark.asyncio
async def test_connection_validation_seqnum_toolow_poss_dup_flag(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=None, connection=conn)

    conn._session.next_num_out = 20
    conn._connection_state = ConnectionState.RESENDREQ_AWAITING

    msg = FIXMessage(FMsg.NEWS, {FTag.PossDupFlag: "Y", FTag.MsgSeqNum: 19})
    await conn.send_msg(msg)
    msg_out = ft.initiator_sent[-1]
    ft.set_next_num(num_in=21)

    ft.conn_accept._connection_state = ConnectionState.RESENDREQ_AWAITING
    assert msg_out[FTag.PossDupFlag] == "Y"
    assert ft.conn_accept._validate_integrity(msg_out) is None


@pytest.mark.asyncio
async def test_connection_validation_beginstring(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    msg_out = ft.initiator_sent[-1]
    msg_out.set(FTag.BeginString, "FIX4.8", replace=True)

    assert (
        ft.conn_accept._validate_integrity(msg_out)
        == "Protocol BeginString(8) mismatch, expected FIX.4.4, got FIX4.8"
    )


@pytest.mark.asyncio
async def test_connection_validation_no_target(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    msg_out = ft.initiator_sent[-1]
    del msg_out[FTag.TargetCompID]

    assert conn._validate_integrity(msg_out) is True


@pytest.mark.asyncio
async def test_connection_validation_no_sender(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    msg_out = ft.initiator_sent[-1]
    del msg_out[FTag.SenderCompID]

    assert conn._validate_integrity(msg_out) is True


@pytest.mark.asyncio
async def test_connection_validation_sender_mismatch(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    msg_out = ft.initiator_sent[-1]
    msg_out.set(FTag.SenderCompID, "as", replace=True)

    assert conn._validate_integrity(msg_out) == "TargetCompID / SenderCompID mismatch"


@pytest.mark.asyncio
async def test_connection_validation_target_mismatch(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    msg_out = ft.initiator_sent[-1]
    msg_out.set(FTag.TargetCompID, "as", replace=True)

    assert conn._validate_integrity(msg_out) == "TargetCompID / SenderCompID mismatch"


@pytest.mark.asyncio
async def test_connection__process_resend_req_synth(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    with patch.object(conn, "_journaler") as mock__journaler:
        msgs = [
            FIXMessage(FMsg.LOGON),
            FIXMessage(FMsg.HEARTBEAT),
            FIXNewOrderSingle("test", "ticker", "1", 10, 10).new_req(),
            FIXMessage(FMsg.HEARTBEAT),
            FIXMessage(FMsg.TESTREQUEST),
            FIXMessage(FMsg.RESENDREQUEST),
            FIXMessage(FMsg.SEQUENCERESET, {FTag.NewSeqNo: 800}),
            FIXNewOrderSingle("test", "ticker", "1", 10, 10).new_req(),
            FIXMessage(FMsg.TESTREQUEST),
        ]
        seq_res = msgs[6]
        assert seq_res.msg_type == FMsg.SEQUENCERESET
        seq_res[34] = conn._session.next_num_out + 5
        assert seq_res[34] == "6"
        enc_msg = [conn._codec.encode(m, conn._session).encode() for m in msgs]
        mock__journaler.recover_messages.return_value = enc_msg

        conn._session.next_num_out = 20

        resend_req = ft.msg_resend_request(1, end_seq_no=0)
        conn._connection_state = ConnectionState.RESENDREQ_HANDLING
        await conn._process_resend(resend_req)
        conn._connection_state = ConnectionState.ACTIVE

        assert len(ft.initiator_sent) == 5
        assert ft.initiator_sent[0].query(35, 34, 36, 123) == {
            FTag.MsgType: str(FMsg.SEQUENCERESET),
            FTag.MsgSeqNum: "1",
            FTag.NewSeqNo: "3",
            FTag.GapFillFlag: "Y",
        }

        assert ft.initiator_sent[1].query(35, 34, FTag.PossDupFlag) == {
            FTag.MsgType: str(FMsg.NEWORDERSINGLE),
            FTag.MsgSeqNum: "3",
            FTag.PossDupFlag: "Y",
        }

        assert ft.initiator_sent[2].query(35, 34, 36, 123) == {
            FTag.MsgType: str(FMsg.SEQUENCERESET),
            FTag.MsgSeqNum: "4",
            FTag.NewSeqNo: "7",
            FTag.GapFillFlag: "Y",
        }

        assert ft.initiator_sent[3].query(35, 34, FTag.PossDupFlag) == {
            FTag.MsgType: str(FMsg.NEWORDERSINGLE),
            FTag.MsgSeqNum: "7",
            FTag.PossDupFlag: "Y",
        }

        assert ft.initiator_sent[4].query(35, 34, 36, 123) == {
            FTag.MsgType: str(FMsg.SEQUENCERESET),
            FTag.MsgSeqNum: "8",
            FTag.NewSeqNo: "20",
            FTag.GapFillFlag: "Y",
        }


@pytest.mark.asyncio
async def test_sequence_reset_request__incoming_seq_num_toolow_ignored(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    await ft.process_msg_acceptor()

    assert ft.conn_accept._connection_state == ConnectionState.ACTIVE
    assert ft.conn_init._connection_state == ConnectionState.ACTIVE
    assert ft.conn_init._session.next_num_out == 2
    ft.conn_accept._session.next_num_in = 20

    seqreset_msg = FIXMessage(
        FMsg.SEQUENCERESET,
        {FTag.NewSeqNo: 10, FTag.MsgSeqNum: conn._session.next_num_out},
    )
    await conn.send_msg(seqreset_msg)
    with patch.object(ft.conn_accept._journaler, "set_seq_num") as mock_set_seq_num:
        await ft.process_msg_acceptor()
        assert mock_set_seq_num.call_count == 2
        assert mock_set_seq_num.call_args_list[0][0] == (ft.conn_accept._session,)
        assert mock_set_seq_num.call_args_list[0][1] == {"next_num_in": 2}

        assert mock_set_seq_num.call_args_list[1][0] == (ft.conn_accept._session,)
        assert mock_set_seq_num.call_args_list[1][1] == {"next_num_in": 10}

    assert ft.conn_accept._session.next_num_in == 10


@pytest.mark.asyncio
async def test_sequence_reset_request__no_gap(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    await ft.process_msg_acceptor()

    assert ft.conn_accept._connection_state == ConnectionState.ACTIVE
    assert ft.conn_init._connection_state == ConnectionState.ACTIVE
    assert ft.conn_init._session.next_num_out == 2

    seqreset_msg = FIXMessage(
        FMsg.SEQUENCERESET,
        {FTag.NewSeqNo: 10, FTag.MsgSeqNum: conn._session.next_num_out},
    )
    await conn.send_msg(seqreset_msg)
    with patch.object(ft.conn_accept._journaler, "set_seq_num") as mock_set_seq_num:
        await ft.process_msg_acceptor()
        assert mock_set_seq_num.call_count == 2
        assert mock_set_seq_num.call_args_list[0][0] == (ft.conn_accept._session,)
        assert mock_set_seq_num.call_args_list[0][1] == {"next_num_in": 2}

        assert mock_set_seq_num.call_args_list[1][0] == (ft.conn_accept._session,)
        assert mock_set_seq_num.call_args_list[1][1] == {"next_num_in": 10}

    assert ft.conn_accept._session.next_num_in == 10


@pytest.mark.asyncio
async def test_sequence_reset_request__unexpected_gapfillflag(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    await ft.process_msg_acceptor()

    assert ft.conn_accept._connection_state == ConnectionState.ACTIVE
    assert ft.conn_init._connection_state == ConnectionState.ACTIVE

    seqreset_msg = FIXMessage(
        FMsg.SEQUENCERESET,
        {
            FTag.NewSeqNo: 10,
            FTag.MsgSeqNum: conn._session.next_num_out,
            FTag.GapFillFlag: "Y",
        },
    )
    await conn.send_msg(seqreset_msg)
    await ft.process_msg_acceptor()

    assert ft.conn_accept._session.next_num_in == 10


@pytest.mark.asyncio
async def test__finalize_message(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    msg = ft.msg_sequence_reset(1, 10, is_gap_fill=True)
    assert conn._message_last_time == 0

    with patch.object(conn, "_journaler") as mock_journaler:
        await conn._finalize_message(msg, b"msg")
        assert conn._message_last_time > 0
        assert mock_journaler.persist_msg.called
        assert mock_journaler.persist_msg.call_args[0] == (
            b"msg",
            conn._session,
            MessageDirection.INBOUND,
        )
        assert mock_journaler.persist_msg.call_args[1] == {}
        assert conn._session.next_num_in == 10

    msg = ft.msg_sequence_reset(10, 15, is_gap_fill=True)
    del msg[FTag.NewSeqNo]
    assert conn._session.set_next_num_in(msg) == 0

    with patch.object(conn, "_journaler") as mock_journaler:
        await conn._finalize_message(msg, b"msg")
        assert not mock_journaler.persist_msg.called
        assert conn._session.next_num_in == 10

    msg = ft.msg_logon()
    assert FTag.MsgSeqNum not in msg
    assert conn._session.set_next_num_in(msg) == 0

    with patch.object(conn, "_journaler") as mock_journaler:
        await conn._finalize_message(msg, b"msg")
        assert not mock_journaler.persist_msg.called
        assert conn._session.next_num_in == 10

    msg = ft.msg_logon()
    msg[FTag.MsgSeqNum] = 100
    assert conn._session.set_next_num_in(msg) == -1

    with patch.object(conn, "_journaler") as mock_journaler:
        await conn._finalize_message(msg, b"msg")
        assert not mock_journaler.persist_msg.called
        assert conn._session.next_num_in == 10


@pytest.mark.asyncio
async def test_connection_both_seqnum_mismach_bidirectional_resend_req(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    conn._session.next_num_out = 20
    conn._session.next_num_in = 25
    ft.set_next_num(num_in=15, num_out=30)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    await ft.process_msg_acceptor()
    assert ft.conn_init._connection_state == ConnectionState.ACTIVE
    assert ft.conn_accept._connection_state == ConnectionState.ACTIVE

    conn.log.debug(100 * "-")
    conn._session.next_num_out = 40
    await conn.send_msg(ft.msg_heartbeat())
    await ft.process_msg_acceptor()
    assert ft.conn_init._connection_state == ConnectionState.ACTIVE
    assert ft.conn_accept._connection_state == ConnectionState.ACTIVE


@pytest.mark.asyncio
async def test_test_request_exchange(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    await ft.process_msg_acceptor()
    assert ft.conn_accept._connection_state == ConnectionState.ACTIVE
    assert ft.conn_init._connection_state == ConnectionState.ACTIVE

    await conn.send_test_req()
    await ft.process_msg_acceptor()

    assert conn._test_req_id is None
    assert ft.conn_accept._connection_state == ConnectionState.ACTIVE
    assert ft.conn_init._connection_state == ConnectionState.ACTIVE


@pytest.mark.asyncio
async def test_test_incorrect_response(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    await ft.process_msg_acceptor()
    assert ft.conn_accept._connection_state == ConnectionState.ACTIVE
    assert ft.conn_init._connection_state == ConnectionState.ACTIVE

    with patch.object(ft.conn_accept, "_process_testrequest") as mock__process_testreq:

        async def _mock_test_req(msg):
            m = ft.msg_heartbeat("asd")
            await ft.conn_accept.send_msg(m)

        mock__process_testreq.side_effect = _mock_test_req

        await conn.send_test_req()
        await ft.process_msg_acceptor()

        assert conn._test_req_id is None
        assert (
            ft.conn_accept._connection_state == ConnectionState.DISCONNECTED_WCONN_TODAY
        )
        assert (
            ft.conn_init._connection_state == ConnectionState.DISCONNECTED_BROKEN_CONN
        )


@pytest.mark.asyncio
async def test_test_request_errors_side_cases(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    await ft.process_msg_acceptor()
    assert ft.conn_accept._connection_state == ConnectionState.ACTIVE
    assert ft.conn_init._connection_state == ConnectionState.ACTIVE

    with pytest.raises(
        FIXConnectionError,
        match=r"You must rend TestRequest\(\) message via self.send_test_req",
    ):
        testmsg = ft.msg_test_request("123")
        await conn.send_msg(testmsg)

    with patch.object(ft.conn_accept, "_process_testrequest") as mock__process_testreq:

        async def _mock_test_req(msg):
            m = ft.msg_heartbeat()
            await ft.conn_accept.send_msg(m)

        mock__process_testreq.side_effect = _mock_test_req

        await conn.send_test_req()
        await ft.process_msg_acceptor()

        with pytest.raises(
            FIXConnectionError, match=r"Another test request already pending"
        ):
            await conn.send_test_req()

        # No Heartbeat(TestReqId) given, just skip until we get valid one
        assert conn._test_req_id is not None
        assert ft.conn_accept._connection_state == ConnectionState.ACTIVE
        assert ft.conn_init._connection_state == ConnectionState.ACTIVE


@pytest.mark.asyncio
async def test_extra_msg_during_seq_num_resend_alredy_journaled(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    conn._session.next_num_out = 20
    conn._session.next_num_in = 25
    ft.set_next_num(num_in=15, num_out=30)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    await ft.process_msg_acceptor()
    assert ft.conn_init._connection_state == ConnectionState.ACTIVE
    assert ft.conn_accept._connection_state == ConnectionState.ACTIVE

    conn.log.debug(100 * "-")

    # replace INITIATOR Journaler to avoid duplicate SQL error (just mock)
    conn._journaler = Journaler()
    conn._session.next_num_out = 15

    ord = FIXNewOrderSingle("test", "ticker", "1", 10, 10).new_req()
    ord[FTag.MsgSeqNum] = 15
    ord[FTag.PossDupFlag] = "Y"

    await conn.send_msg(ord)
    await ft.process_msg_acceptor()

    assert ft.conn_init._connection_state == ConnectionState.DISCONNECTED_WCONN_TODAY
    assert ft.conn_accept._connection_state == ConnectionState.DISCONNECTED_BROKEN_CONN


@pytest.mark.asyncio
async def test_extra_msg_during_seq_num_resend_low_num_not_processed(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    conn._session.next_num_out = 20
    conn._session.next_num_in = 25
    ft.set_next_num(num_in=15, num_out=30)

    msg = ft.msg_logon()
    await conn.send_msg(msg)
    await ft.process_msg_acceptor()
    assert ft.conn_init._connection_state == ConnectionState.ACTIVE
    assert ft.conn_accept._connection_state == ConnectionState.ACTIVE

    conn.log.debug(100 * "-")
    # ft.set_next_num(num_in=15)
    with patch.object(ft.conn_accept, "_journaler") as mock_journaler:
        mock_journaler.persist_msg.reset_mock()
        conn._session.next_num_out = 17
        await conn.send_msg(ft.msg_heartbeat())
        await conn.send_msg(ft.msg_heartbeat())
        await ft.process_msg_acceptor()

        assert (
            ft.conn_init._connection_state == ConnectionState.DISCONNECTED_WCONN_TODAY
        )
        assert (
            ft.conn_accept._connection_state == ConnectionState.DISCONNECTED_BROKEN_CONN
        )
        assert len(mock_journaler.persist_msg.call_args_list) == 1
        assert (
            mock_journaler.persist_msg.call_args_list[0][0][2]
            == MessageDirection.OUTBOUND
        )


@pytest.mark.asyncio
async def test_connection__process_resend_req_real_processing(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    with (
        patch.object(conn, "_journaler") as mock__journaler,
        patch.object(ft.conn_accept, "on_message") as mock_on_message,
    ):
        msgs = [
            FIXMessage(FMsg.LOGON),
            FIXMessage(FMsg.HEARTBEAT),
            FIXNewOrderSingle("test", "ticker", "1", 10, 10).new_req(),
            FIXMessage(FMsg.HEARTBEAT),
            FIXMessage(FMsg.TESTREQUEST),
            FIXMessage(FMsg.RESENDREQUEST),
        ]
        enc_msg = [conn._codec.encode(m, conn._session).encode() for m in msgs]
        mock__journaler.recover_messages.return_value = enc_msg
        conn._session.next_num_out = 10
        msg = ft.msg_logon()
        await conn.send_msg(msg)
        await ft.process_msg_acceptor()
        assert ft.conn_init._connection_state == ConnectionState.ACTIVE
        assert ft.conn_accept._connection_state == ConnectionState.ACTIVE
        assert mock_on_message.called
        assert mock_on_message.call_count == 1
        (app_msg,) = mock_on_message.call_args[0]
        assert app_msg.query(35) == {FTag.MsgType: FMsg.NEWORDERSINGLE}


@pytest.mark.asyncio
async def test_connection__process_resend__ignores_high_seq_num_msg(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    conn._session.next_num_out = 10

    with (
        patch.object(conn, "_journaler") as mock__journaler,
        patch.object(ft.conn_accept, "on_message") as mock_on_message,
    ):
        msgs = [
            FIXMessage(FMsg.LOGON),
            FIXMessage(FMsg.HEARTBEAT),
            FIXNewOrderSingle("test", "ticker", "1", 10, 10).new_req(),
            FIXMessage(FMsg.HEARTBEAT),
            FIXMessage(FMsg.TESTREQUEST),
            FIXMessage(FMsg.RESENDREQUEST),
            FIXNewOrderSingle("test2", "ticker", "1", 10, 10).new_req(),
        ]
        enc_msg = [conn._codec.encode(m, conn._session).encode() for m in msgs]
        mock__journaler.recover_messages.return_value = enc_msg
        msg = ft.msg_logon()
        await conn.send_msg(msg)

        await ft.process_msg_acceptor(0)
        conn.log.debug(100 * "-")
        assert ft.conn_accept._connection_state == ConnectionState.RESENDREQ_AWAITING
        # Pretending that last order was sent via wire, and received before RESENDREQUEST
        #  we are sending already encoded message via socket mock
        dec_msg, _, _ = conn._codec.decode(enc_msg[-1])
        assert isinstance(dec_msg, FIXMessage)
        assert dec_msg[35] == FMsg.NEWORDERSINGLE

        # Prepend que to make OrderSingle arrive first
        ft.acceptor_rcv_que.insert(0, (dec_msg, enc_msg[-1]))
        await ft.process_msg_acceptor()

        assert mock_on_message.called
        assert mock_on_message.call_count == 2
        (app_msg,) = mock_on_message.call_args_list[0][0]
        assert app_msg.query(35, FTag.ClOrdID, FTag.PossDupFlag) == {
            FTag.MsgType: FMsg.NEWORDERSINGLE,
            FTag.ClOrdID: "test--1",
            FTag.PossDupFlag: "Y",
        }
        (app_msg,) = mock_on_message.call_args_list[1][0]
        assert app_msg.query(35, FTag.ClOrdID, FTag.PossDupFlag) == {
            FTag.MsgType: FMsg.NEWORDERSINGLE,
            FTag.ClOrdID: "test2--1",
            FTag.PossDupFlag: "Y",
        }


@pytest.mark.asyncio
async def test_connection_init_launch_tasks(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    journaler_mock = MagicMock()
    with (
        patch("asyncio.create_task") as mock_create_task,
        patch("asyncfix.connection.AsyncFIXConnection.socket_read_task") as t1,
        patch("asyncfix.connection.AsyncFIXConnection.heartbeat_timer_task") as t2,
    ):
        connection = AsyncFIXConnection(
            FIXProtocol44(),
            "INITIATOR",
            "ACCEPTOR",
            journaler=journaler_mock,
            host="localhost",
            port="64444",
            heartbeat_period=33,
        )

        assert connection._journaler is journaler_mock
        assert journaler_mock.create_or_load.call_args[0] == ("ACCEPTOR", "INITIATOR")
        assert journaler_mock.create_or_load.call_args[1] == {}

        assert connection._session == journaler_mock.create_or_load()
        assert connection._connection_state == ConnectionState.DISCONNECTED_NOCONN_TODAY
        assert connection.connection_role == ConnectionRole.UNKNOWN
        assert isinstance(connection._codec.protocol, FIXProtocol44)
        assert not connection._connection_was_active
        assert connection.heartbeat_period == 33
        assert connection._host == "localhost"
        assert connection._port == 64444
        assert connection._socket_reader is None
        assert connection._socket_writer is None
        assert connection._message_last_time == 0
        assert connection._max_seq_num_resend == 0
        assert not connection._msg_buffer
        assert connection._test_req_id is None

        assert mock_create_task.call_count == 0

        assert connection._aio_task_heartbeat is None
        assert connection._aio_task_socket_read is None
        await connection.connect()
        assert connection._aio_task_heartbeat is not None
        assert connection._aio_task_socket_read is not None
        assert t1.called
        assert t2.called

        # This one just suppress warning about coro was not awaited
        await mock_create_task.call_args_list[0][0][0]
        await mock_create_task.call_args_list[1][0][0]

        # Next connect doesn't create tasks again
        t1.reset_mock()
        t2.reset_mock()
        await connection.connect()
        assert not t1.called
        assert not t2.called
        assert connection._aio_task_heartbeat is not None
        assert connection._aio_task_socket_read is not None


@pytest.mark.asyncio
async def test_connect_raises(fix_connection):
    conn: AsyncFIXConnection = fix_connection

    with pytest.raises(
        NotImplementedError, match=r"on_connect\(\) must be implemented in app class"
    ):
        await conn.on_connect()

    with pytest.raises(
        NotImplementedError, match=r"on_message\(\) must be implemented in app class"
    ):
        await conn.on_message(None)


@pytest.mark.asyncio
async def test_process_message_exception(fix_connection):
    conn: AsyncFIXConnection = fix_connection

    conn.log = MagicMock()
    with (
        patch.object(conn, "_validate_integrity") as mock__validate_integrity,
        patch.object(conn, "_finalize_message") as mock_finalize_message,
        patch.object(conn, "_process_logon") as mock_process_logon,
    ):
        mock__validate_integrity.side_effect = RuntimeError

        msg = FIXMessage(FMsg.LOGON)

        with pytest.raises(RuntimeError):
            await conn._process_message(msg, b"msg")

        mock__validate_integrity.return_value = None
        mock__validate_integrity.side_effect = None
        mock_process_logon.side_effect = RuntimeError
        await conn._process_message(msg, b"msg")

        # except block
        assert conn.log.exception.called

        # finally block
        #  not called because is_valid_msg_num - not passed
        assert not mock_finalize_message.called

        mock_process_logon.side_effect = asyncio.CancelledError
        with pytest.raises(asyncio.CancelledError):
            await conn._process_message(msg, b"msg")


@pytest.mark.asyncio
async def test_socket_read__valid(fix_connection_socket, fix_msg):
    conn: AsyncFIXConnection = fix_connection_socket
    socket_reader_mock = conn._socket_reader

    socket_reader_mock.read.side_effect = [
        conn._codec.encode(fix_msg, conn._session).encode(),
        conn._codec.encode(fix_msg, conn._session).encode(),
        asyncio.CancelledError,  # This stops infinite loop
    ]

    with patch.object(conn, "_process_message") as mock_process_message:
        await asyncio.wait_for(conn.socket_read_task(), 1)

    assert mock_process_message.call_count == 2

    for args in mock_process_message.call_args_list:
        assert len(args[0]) == 2  # 2 args
        assert args[1] == {}  # no kwargs

        assert {
            FTag.Symbol: "VOD.L",
            FTag.Price: "123.45",
            FTag.OrderQty: "9876",
        } == args[0][0].query(FTag.Symbol, FTag.Price, FTag.OrderQty)

        assert isinstance(args[0][1], bytes)
        assert args[0][1].startswith(b"8=FIX")


@pytest.mark.asyncio
async def test_socket_read__partial(fix_connection_socket, fix_msg):
    conn: AsyncFIXConnection = fix_connection_socket
    socket_reader_mock = conn._socket_reader

    full = conn._codec.encode(fix_msg, conn._session).encode()
    partial_1 = full[:20]
    partial_2 = full[20:]
    assert partial_1 + partial_2 == full

    socket_reader_mock.read.side_effect = [
        conn._codec.encode(fix_msg, conn._session).encode(),
        partial_1,
        partial_2,
        asyncio.CancelledError,  # This stops infinite loop
    ]

    with patch.object(conn, "_process_message") as mock_process_message:
        await asyncio.wait_for(conn.socket_read_task(), 1)

    for args in mock_process_message.call_args_list:
        assert len(args[0]) == 2  # 2 args
        assert args[1] == {}  # no kwargs

        assert {
            FTag.Symbol: "VOD.L",
            FTag.Price: "123.45",
            FTag.OrderQty: "9876",
        } == args[0][0].query(FTag.Symbol, FTag.Price, FTag.OrderQty)

        assert isinstance(args[0][1], bytes)
        assert args[0][1].startswith(b"8=FIX")


@pytest.mark.asyncio
async def test_socket_read__partial_bad_data(fix_connection_socket, fix_msg):
    conn: AsyncFIXConnection = fix_connection_socket
    socket_reader_mock = conn._socket_reader
    full = conn._codec.encode(fix_msg, conn._session).encode()
    partial_1 = full[:20]
    partial_2 = full[20:]
    assert partial_1 + partial_2 == full

    socket_reader_mock.read.side_effect = [
        conn._codec.encode(fix_msg, conn._session).encode(),
        partial_1,
        partial_1,
        asyncio.CancelledError,  # This stops infinite loop
    ]

    with patch.object(conn, "_process_message") as mock_process_message:
        await asyncio.wait_for(conn.socket_read_task(), 1)

    assert mock_process_message.call_count == 1

    for args in mock_process_message.call_args_list:
        assert len(args[0]) == 2  # 2 args
        assert args[1] == {}  # no kwargs

        assert {
            FTag.Symbol: "VOD.L",
            FTag.Price: "123.45",
            FTag.OrderQty: "9876",
        } == args[0][0].query(FTag.Symbol, FTag.Price, FTag.OrderQty)

        assert isinstance(args[0][1], bytes)
        assert args[0][1].startswith(b"8=FIX")


@pytest.mark.asyncio
async def test_socket_read__decode_logic_msg_buff_management(
    fix_connection_socket, fix_msg
):
    conn: AsyncFIXConnection = fix_connection_socket
    socket_reader_mock = conn._socket_reader

    full = conn._codec.encode(fix_msg, conn._session).encode()
    partial_1 = full[:20]
    partial_2 = full[20:]
    assert partial_1 + partial_2 == full

    msg_valid = b"8=FIX.4.4\x019=82\x0135=D\x0149=sender\x0156=target\x0134=1\x0152=20230919-07:13:26.808\x0144=123.45\x0138=9876\x0155=VOD.L\x0110=100\x01"  # noqa

    with patch.object(conn, "_process_message") as mock_process_message:
        socket_reader_mock.read.side_effect = [
            b"abcd",
            msg_valid,
            asyncio.CancelledError,  # This stops infinite loop
        ]
        await asyncio.wait_for(conn.socket_read_task(), 1)

    assert mock_process_message.call_count == 1


@pytest.mark.asyncio
async def test_socket_read__partial_one_after_another(fix_connection_socket, fix_msg):
    conn: AsyncFIXConnection = fix_connection_socket
    socket_reader_mock = conn._socket_reader
    full = conn._codec.encode(fix_msg, conn._session).encode()
    partial_1 = full[:20]
    partial_2 = full[20:]
    assert partial_1 + partial_2 == full

    socket_reader_mock.read.side_effect = [
        partial_1,
        # remainder comes with next valid message
        partial_2 + conn._codec.encode(fix_msg, conn._session).encode(),
        asyncio.CancelledError,  # This stops infinite loop
    ]

    with patch.object(conn, "_process_message") as mock_process_message:
        await asyncio.wait_for(conn.socket_read_task(), 1)

    assert mock_process_message.call_count == 2

    for args in mock_process_message.call_args_list:
        assert len(args[0]) == 2  # 2 args
        assert args[1] == {}  # no kwargs

        assert {
            FTag.Symbol: "VOD.L",
            FTag.Price: "123.45",
            FTag.OrderQty: "9876",
        } == args[0][0].query(FTag.Symbol, FTag.Price, FTag.OrderQty)

        assert isinstance(args[0][1], bytes)
        assert args[0][1].startswith(b"8=FIX")


@pytest.mark.asyncio
async def test_socket_read__connection_closed_by_peer(fix_connection_socket, fix_msg):
    conn: AsyncFIXConnection = fix_connection_socket
    socket_reader_mock = conn._socket_reader

    conn._connection_state = ConnectionState.ACTIVE

    socket_reader_mock.read.side_effect = [
        ConnectionError,
        asyncio.CancelledError,  # This stops infinite loop
    ]
    mock_socket_writer = conn._socket_writer
    mock_socket_reader = conn._socket_reader

    with (
        patch.object(conn, "_process_message") as mock_process_message,
        patch("asyncio.sleep") as mock_sleep,
        patch.object(conn, "send_msg") as mock_send_msg,
    ):
        mock_sleep.side_effect = [
            1,
            asyncio.CancelledError,
        ]
        await asyncio.wait_for(conn.socket_read_task(), 1)

    assert conn.connection_state == ConnectionState.DISCONNECTED_BROKEN_CONN
    assert conn._socket_reader is None
    assert conn._socket_writer is None
    assert mock_socket_writer.close.called
    assert mock_socket_writer.wait_closed.awaited
    assert not mock_send_msg.called  # no LOGOUT sent!

    assert mock_sleep.call_count == 2
    assert mock_sleep.call_args[0] == (1,)


@pytest.mark.asyncio
async def test_socket_read__empty_message_closes_socket(fix_connection_socket, fix_msg):
    conn: AsyncFIXConnection = fix_connection_socket
    socket_reader_mock = conn._socket_reader

    conn._connection_state = ConnectionState.ACTIVE

    socket_reader_mock.read.side_effect = [
        b"",
        asyncio.CancelledError,  # This stops infinite loop
    ]
    mock_socket_writer = conn._socket_writer
    mock_socket_reader = conn._socket_reader

    with (
        patch.object(conn, "_process_message") as mock_process_message,
        patch("asyncio.sleep") as mock_sleep,
        patch.object(conn, "send_msg") as mock_send_msg,
    ):
        mock_sleep.side_effect = [
            1,
            asyncio.CancelledError,
        ]
        await asyncio.wait_for(conn.socket_read_task(), 1)

    assert conn.connection_state == ConnectionState.DISCONNECTED_BROKEN_CONN
    assert conn._socket_reader is None
    assert conn._socket_writer is None
    assert mock_socket_writer.close.called
    assert mock_socket_writer.wait_closed.awaited
    assert not mock_send_msg.called  # no LOGOUT sent!

    assert mock_sleep.call_count == 2
    assert mock_sleep.call_args[0] == (1,)


@pytest.mark.asyncio
async def test_socket_read__exception_inside(fix_connection_socket, fix_msg):
    conn: AsyncFIXConnection = fix_connection_socket
    socket_reader_mock = conn._socket_reader

    conn._connection_state = ConnectionState.ACTIVE

    socket_reader_mock.read.side_effect = [
        conn._codec.encode(fix_msg, conn._session).encode(),
        conn._codec.encode(fix_msg, conn._session).encode(),
        asyncio.CancelledError,  # This stops infinite loop
    ]
    mock_socket_writer = conn._socket_writer
    mock_socket_reader = conn._socket_reader

    def _process_side(*args):
        conn._connection_state = ConnectionState.DISCONNECTED_BROKEN_CONN
        raise RuntimeError("test")

    with (
        patch.object(conn, "_process_message") as mock_process_message,
        patch("asyncio.sleep") as mock_sleep,
        patch.object(conn, "send_msg") as mock_send_msg,
    ):
        mock_process_message.side_effect = _process_side
        await asyncio.wait_for(conn.socket_read_task(), 1)

    assert mock_process_message.call_count == 1
    assert conn.connection_state == ConnectionState.DISCONNECTED_BROKEN_CONN
    assert conn._socket_reader is not None
    assert conn._socket_writer is not None

    assert not mock_sleep.called


@pytest.mark.asyncio
async def test_disconnect(fix_connection_socket, fix_msg):
    conn: AsyncFIXConnection = fix_connection_socket
    socket_reader_mock = conn._socket_reader

    conn._connection_state = ConnectionState.AWAITING_CONNECTION
    with (
        patch.object(conn, "_process_message") as mock_process_message,
        patch("asyncio.sleep") as mock_sleep,
        patch.object(conn, "send_msg") as mock_send_msg,
    ):
        mock_socket_writer = conn._socket_writer
        mock_socket_reader = conn._socket_reader
        conn._test_req_id = 123
        conn._message_last_time = 222
        conn._max_seq_num_resend = 8

        await conn.disconnect(
            ConnectionState.DISCONNECTED_WCONN_TODAY, logout_message="test disconn"
        )

        assert conn.connection_state == ConnectionState.DISCONNECTED_WCONN_TODAY
        assert conn._socket_reader is None
        assert conn._socket_writer is None
        assert mock_socket_writer.close.called
        assert mock_socket_writer.wait_closed.awaited
        assert mock_send_msg.call_count == 1

        assert isinstance(mock_send_msg.call_args[0][0], FIXMessage)
        assert mock_send_msg.call_args[0][0].msg_type == FMsg.LOGOUT
        assert mock_send_msg.call_args[0][0].query(FTag.Text) == {
            FTag.Text: "test disconn",
        }

        assert conn._test_req_id is None
        assert conn._message_last_time == 0
        assert conn._max_seq_num_resend == 0

        # already disconnected, skipped
        mock_send_msg.reset_mock()
        await conn.disconnect(
            ConnectionState.DISCONNECTED_WCONN_TODAY, logout_message="test disconn"
        )
        assert mock_send_msg.call_count == 0


@pytest.mark.asyncio
async def test_disconnect_empty_text_logout(fix_connection_socket, fix_msg):
    conn: AsyncFIXConnection = fix_connection_socket
    socket_reader_mock = conn._socket_reader

    conn._connection_state = ConnectionState.AWAITING_CONNECTION
    with (
        patch.object(conn, "_process_message") as mock_process_message,
        patch("asyncio.sleep") as mock_sleep,
        patch.object(conn, "send_msg") as mock_send_msg,
    ):
        await conn.disconnect(
            ConnectionState.DISCONNECTED_WCONN_TODAY, logout_message=""
        )

        assert isinstance(mock_send_msg.call_args[0][0], FIXMessage)
        assert mock_send_msg.call_args[0][0].msg_type == FMsg.LOGOUT
        assert FTag.Text not in mock_send_msg.call_args[0][0]


@pytest.mark.asyncio
async def test_heartbeat_task__notconnected_socket(fix_connection_socket, fix_msg):
    conn: AsyncFIXConnection = fix_connection_socket
    conn._connection_state = ConnectionState.AWAITING_CONNECTION
    with (
        patch.object(conn, "disconnect") as mock_disconnect,
        patch("asyncio.sleep") as mock_sleep,
        patch.object(conn, "send_test_req") as mock_send_test_req,
    ):
        mock_sleep.side_effect = [1, asyncio.CancelledError]

        conn._socket_reader = None
        await conn.heartbeat_timer_task()

        assert mock_sleep.call_count == 2
        assert mock_sleep.call_args[0] == (1,)

        assert not mock_disconnect.called
        assert not mock_send_test_req.called


@pytest.mark.asyncio
async def test_socket_read_task__socket_reconnect(fix_connection_socket, fix_msg):
    conn: AsyncFIXConnection = fix_connection_socket
    conn._connection_state = ConnectionState.AWAITING_CONNECTION
    conn._connection_role = ConnectionRole.INITIATOR

    with (
        patch.object(conn, "disconnect") as mock_disconnect,
        patch("asyncio.sleep") as mock_sleep,
        patch.object(conn, "connect") as mock_connect,
    ):
        mock_sleep.side_effect = [1, asyncio.CancelledError]

        conn._socket_reader = None
        conn._heartbeat_period = -1  # let connect call immediately
        await conn.socket_read_task()

        assert mock_sleep.call_count == 2
        assert mock_sleep.call_args[0] == (1,)

        assert not mock_disconnect.called
        assert mock_connect.await_count == 1


@pytest.mark.asyncio
async def test_heartbeat_task__idle_w_exception(fix_connection_socket, fix_msg):
    conn: AsyncFIXConnection = fix_connection_socket
    conn._connection_state = ConnectionState.AWAITING_CONNECTION
    with (
        patch.object(conn, "disconnect") as mock_disconnect,
        patch("asyncio.sleep") as mock_sleep,
        patch.object(conn, "send_test_req") as mock_send_test_req,
    ):
        mock_sleep.side_effect = [1, RuntimeError, asyncio.CancelledError]

        conn._heartbeat_period = -1  # Negative always triggers
        conn._test_req_id = None
        conn._message_last_time = 0.0

        await conn.heartbeat_timer_task()

        assert mock_sleep.call_count == 3

        assert not mock_disconnect.called
        assert not mock_send_test_req.called


@pytest.mark.asyncio
async def test_heartbeat_task__msg_last_time_disconnect(fix_connection_socket, fix_msg):
    conn: AsyncFIXConnection = fix_connection_socket
    conn._connection_state = ConnectionState.AWAITING_CONNECTION
    with (
        patch.object(conn, "disconnect") as mock_disconnect,
        patch("asyncio.sleep") as mock_sleep,
        patch.object(conn, "send_test_req") as mock_send_test_req,
    ):
        mock_sleep.side_effect = [1, 1, asyncio.CancelledError]

        conn._heartbeat_period = -1  # Negative always triggers
        conn._test_req_id = None
        conn._message_last_time = 1.0

        def _disconn_side(*args):
            conn._message_last_time = 0

        mock_disconnect.side_effect = _disconn_side

        await conn.heartbeat_timer_task()

        assert mock_sleep.call_count == 3

        assert mock_disconnect.call_count == 1
        assert mock_disconnect.call_args[0] == (
            ConnectionState.DISCONNECTED_BROKEN_CONN,
        )
        assert mock_disconnect.call_args[1] == {}

        assert not mock_send_test_req.called


@pytest.mark.asyncio
async def test_heartbeat_task__test_req_timeout(fix_connection_socket, fix_msg):
    conn: AsyncFIXConnection = fix_connection_socket
    conn._connection_state = ConnectionState.AWAITING_CONNECTION
    with (
        patch.object(conn, "disconnect") as mock_disconnect,
        patch("asyncio.sleep") as mock_sleep,
        patch.object(conn, "send_test_req") as mock_send_test_req,
    ):
        mock_sleep.side_effect = [1, 1, asyncio.CancelledError]

        conn._heartbeat_period = -1  # Negative always triggers
        conn._test_req_id = 1
        conn._message_last_time = 0

        def _disconn_side(*args):
            conn._test_req_id = None

        mock_disconnect.side_effect = _disconn_side

        await conn.heartbeat_timer_task()

        assert mock_sleep.call_count == 3

        assert mock_disconnect.call_count == 1
        assert mock_disconnect.call_args[0] == (
            ConnectionState.DISCONNECTED_BROKEN_CONN,
        )
        assert mock_disconnect.call_args[1] == {}

        assert not mock_send_test_req.called


@pytest.mark.asyncio
async def test_heartbeat_task__test_req_sent(fix_connection_socket, fix_msg):
    conn: AsyncFIXConnection = fix_connection_socket
    conn._connection_state = ConnectionState.ACTIVE
    with (
        patch.object(conn, "disconnect") as mock_disconnect,
        patch("asyncio.sleep") as mock_sleep,
        patch.object(conn, "send_test_req") as mock_send_test_req,
    ):
        mock_sleep.side_effect = [1, 1, asyncio.CancelledError]

        conn._heartbeat_period = 1  # Negative always triggers
        conn._test_req_id = None
        conn._message_last_time = 0

        def _test_req_side(*args):
            conn._test_req_id = time.time()

        mock_send_test_req.side_effect = _test_req_side

        await conn.heartbeat_timer_task()
        assert mock_sleep.call_count == 3

        assert mock_disconnect.call_count == 0

        assert mock_send_test_req.call_count == 1
        assert conn._message_last_time != 0


@pytest.mark.asyncio
async def test_process_logout(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    conn._connection_state = ConnectionState.ACTIVE

    with (
        patch.object(conn, "on_logout") as mock_on_logout,
        patch.object(conn, "disconnect") as mock_disconnect,
    ):
        await conn._process_logout(ft.msg_logout())

        assert mock_on_logout.called
        assert mock_disconnect.called


@pytest.mark.asyncio
async def test_reset_seqnum(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)
    conn._session.next_num_in = 3
    conn._session.next_num_out = 5

    with patch.object(conn._journaler, "set_seq_num") as mock_set_seq_num:

        def _side_reset(session, next_num_in, next_num_out):
            assert next_num_in == 1
            assert next_num_out == 1
            session.next_num_in = next_num_in
            session.next_num_out = next_num_out

        mock_set_seq_num.side_effect = _side_reset
        await conn.reset_seq_num()
        assert conn._session.next_num_in == 1
        assert conn._session.next_num_in == 1
        assert mock_set_seq_num.call_count == 1


@pytest.mark.asyncio
async def test_connection__process_resend_req__dupe_journal_seqnum(fix_connection):
    conn: AsyncFIXConnection = fix_connection
    ft = FIXTester(schema=FIX_SCHEMA, connection=conn)

    with patch.object(conn._journaler, "recover_messages") as mock__recover_msg:
        msgs = [
            FIXMessage(FMsg.LOGON),
            FIXMessage(FMsg.HEARTBEAT),
            FIXNewOrderSingle("test", "ticker", "1", 10, 10).new_req(),
            FIXMessage(FMsg.HEARTBEAT),
            FIXMessage(FMsg.TESTREQUEST),
            FIXMessage(FMsg.RESENDREQUEST),
            FIXMessage(FMsg.SEQUENCERESET, {FTag.NewSeqNo: 800}),
            FIXNewOrderSingle("test", "ticker", "1", 10, 10).new_req(),
            FIXMessage(FMsg.TESTREQUEST),
        ]
        seq_res = msgs[6]
        assert seq_res.msg_type == FMsg.SEQUENCERESET
        seq_res[34] = conn._session.next_num_out + 5
        assert seq_res[34] == "6"
        enc_msg = [conn._codec.encode(m, conn._session).encode() for m in msgs]
        mock__recover_msg.return_value = enc_msg

        conn._session.next_num_out = 20

        resend_req = ft.msg_resend_request(1, end_seq_no=0)
        conn._connection_state = ConnectionState.RESENDREQ_HANDLING
        await conn._process_resend(resend_req)
        conn._connection_state = ConnectionState.ACTIVE

        assert len(ft.initiator_sent) == 5
        assert conn._session.next_num_out == 20

        # Requested once again
        ft.reset_messages()
        resend_req = ft.msg_resend_request(1, end_seq_no=0)
        conn._connection_state = ConnectionState.RESENDREQ_HANDLING
        await conn._process_resend(resend_req)
        conn._connection_state = ConnectionState.ACTIVE

        assert len(ft.initiator_sent) == 5
        assert conn._session.next_num_out == 20



"""FIX Initiator (client) connection."""
import asyncio
import logging

from asyncfix.connection import AsyncFIXConnection, ConnectionRole, ConnectionState
from asyncfix.errors import FIXConnectionError
from asyncfix.journaler import Journaler
from asyncfix.protocol import FIXProtocolBase


class AsyncFIXClient(AsyncFIXConnection):
    """Generic FIX client."""

    def __init__(
        self,
        protocol: FIXProtocolBase,
        sender_comp_id: str,
        target_comp_id: str,
        journaler: Journaler,
        host: str,
        port: int,
        heartbeat_period: int = 30,
        logger: logging.Logger | None = None,
    ):
        """Initialization.

        Args:
            protocol: FIX protocol used in codec
            sender_comp_id: client sender_comp_id tag
            target_comp_id: client target_comp_id tag
            journaler: message journaler
            host: fix host
            port: fix port
            heartbeat_period: heartbeat_period in seconds
            logger: custom logger instance
        """
        super().__init__(
            protocol=protocol,
            sender_comp_id=sender_comp_id,
            target_comp_id=target_comp_id,
            journaler=journaler,
            host=host,
            port=port,
            heartbeat_period=heartbeat_period,
            logger=logger,
        )
        self._connection_role = ConnectionRole.INITIATOR

    async def connect(self):
        """Connects to the FIX server and initializes session.

        Raises:
            FIXConnectionError: if already connected
        """
        if self._socket_reader:
            raise FIXConnectionError("Socket already connected")

        # this must be called in order to launch socket_reader/heartbeat tasks
        await super().connect()

        try:
            self._socket_reader, self._socket_writer = await asyncio.open_connection(
                self._host, self._port
            )
        except Exception as exc:
            self.log.info(f"Connection error (waiting reconnect): {repr(exc)}")
            self._connection_state = ConnectionState.DISCONNECTED_BROKEN_CONN
        else:
            self._connection_state = ConnectionState.NETWORK_CONN_ESTABLISHED
            await self.on_connect()



"""AsyncFIX core package."""
from .fixtags import FTag
from .msgtype import FMsg
from .message import FIXMessage
from .connection import AsyncFIXConnection, ConnectionRole, ConnectionState
from .connection_client import AsyncFIXClient
from .connection_server import AsyncFIXDummyServer
from .fix_tester import FIXTester
from .journaler import Journaler



"""FIX Protocol Unit Tester."""
from math import isnan, nan
from unittest.mock import AsyncMock, MagicMock

from asyncfix import FIXMessage, FMsg, FTag
from asyncfix.connection import AsyncFIXConnection, ConnectionState
from asyncfix.journaler import Journaler
from asyncfix.protocol import FIXProtocol44, FIXSchema
from asyncfix.protocol.common import FExecType, FOrdStatus
from asyncfix.protocol.order_single import FIXNewOrderSingle


class FIXTester:
    """FIX protocol tester.

    Attributes:
        registered_orders: registered orders (sent via FIXTester)
        schema: fix schema for validation (optional)
        conn_init: fix connection initiator
        conn_accept: fix virtual acceptor (simulated by FIXTester)
        initiator_sent: messages sent by initiator
        acceptor_rcv_que: raw messages received by (simulated acceptor)
        acceptor_sent: FIXMessages sent by simulated acceptor
    """

    def __init__(
        self,
        schema: FIXSchema | None = None,
        connection: AsyncFIXConnection | None = None,
    ):
        """Initialize FIXTester.

        Args:
            schema: (optional) FIXSchema for validating incoming/outgoing messages
            connection: (optional) fix initiator connection
        """
        self.registered_orders = {}
        self.schema = schema
        self._order_id = 0
        self._exec_id = 10000
        self.conn_init = connection
        self.conn_accept = None
        self.initiator_sent: list[FIXMessage] = []
        """List of fix messages sent by self.connection.send_msg()."""

        self.acceptor_rcv_que: list[tuple(FIXMessage, bytes)] = []

        self.acceptor_sent: list[FIXMessage] = []
        """List of fix messages sent by FIXTester.reply()."""

        if connection:
            assert isinstance(connection, AsyncFIXConnection)
            # target and session swapped! Because we mimic the server
            j = Journaler()
            self.conn_accept = AsyncFIXConnection(
                FIXProtocol44(),
                target_comp_id=self.conn_init._session.sender_comp_id,
                sender_comp_id=self.conn_init._session.target_comp_id,
                journaler=j,
                host="localhost",
                port="64444",
                heartbeat_period=30,
                logger=self.conn_init.log,
            )
            self.conn_accept._connection_state = (
                ConnectionState.NETWORK_CONN_ESTABLISHED
            )
            self.conn_accept._session.next_num_out = connection._session.next_num_in
            self.conn_accept._session.next_num_in = connection._session.next_num_out

            connection._socket_writer = MagicMock()
            connection._socket_writer.write.side_effect = (
                self._conn_socket_write_initiator
            )
            connection._socket_writer.drain = AsyncMock()
            connection._socket_writer.wait_closed = AsyncMock()

            self.conn_accept._socket_writer = MagicMock()
            self.conn_accept._socket_writer.write.side_effect = (
                self._conn_socket_write_acceptor
            )
            self.conn_accept._socket_writer.drain = AsyncMock()
            self.conn_accept._socket_writer.drain.side_effect = (
                self._conn_socket_drain_acceptor
            )
            self._socket_drain_in_coro = None
            self.conn_accept._socket_writer.wait_closed = AsyncMock()

    def set_next_num(self, num_in=None, num_out=None):
        """Set expected session seq nums for simulated acceptor.

        Args:
            num_in: next num in expected by simulated acceptor
            num_out: next num out sent by simulated acceptor
        """
        if num_in is not None:
            assert isinstance(num_in, int)
            assert num_in > 0
            self.conn_accept._session.next_num_in = num_in
        if num_out is not None:
            assert isinstance(num_out, int)
            assert num_out > 0
            self.conn_accept._session.next_num_out = num_out

    def reset_messages(self):
        """Reset messages queues of initiator and acceptor."""
        self.initiator_sent.clear()
        self.acceptor_rcv_que.clear()
        self.acceptor_sent.clear()

    def acceptor_sent_query(
        self,
        tags: tuple[FTag | str | int] | None = None,
        index: int = -1,
    ) -> dict[FTag | str, str]:
        """Query message sent from FIXTester to initiator.

        Args:
            tags: tuple of tag numbers
            index: index of self.acceptor_sent list, default -1

        Returns:
            dict {FTag.MsgSeqNum: "34", "12382": "some value"}
        """
        return self.acceptor_sent[index].query(*tags)

    def initiator_sent_query(
        self,
        tags: tuple[FTag | str | int] | None = None,
        index: int = -1,
    ) -> dict[FTag | str, str]:
        """Query message sent from initiator to FixTester.

        Args:
            tags: tuple of tag numbers
            index: index of self.initiator_sent list, default -1

        Returns:
            dict {FTag.MsgSeqNum: "34", "12382": "some value"}
        """
        return self.initiator_sent[index].query(*tags)

    def _conn_socket_write_initiator(self, data):
        msg, _, _ = self.conn_init._codec.decode(data, silent=False)
        if self.schema:
            self.schema.validate(msg)
        self.initiator_sent.append(msg)
        self.acceptor_rcv_que.append((msg, data))

    def _conn_socket_write_acceptor(self, data):
        msg, _, _ = self.conn_accept._codec.decode(data, silent=False)
        if self.schema:
            self.schema.validate(msg)
        self.acceptor_sent.append(msg)

        self._socket_drain_in_coro = self.conn_init._process_message(msg, data)

    async def _conn_socket_drain_acceptor(self):
        try:
            if self._socket_drain_in_coro:
                await self._socket_drain_in_coro
        finally:
            self._socket_drain_in_coro = None

    async def process_msg_acceptor(self, index=None):
        """Processes messages queued by initiator.send_msg().

        Args:
            index: None - processes all messages in que, number - only one at that index
        """
        assert self.acceptor_rcv_que, "No messages in self.acceptor_rcv_que"

        while self.acceptor_rcv_que:
            (msg, raw) = self.acceptor_rcv_que.pop(0 if index is None else index)
            await self.conn_accept._process_message(msg, raw)
            if index is not None:
                break

    async def reply(self, msg: FIXMessage):
        """Manually reply to the initiator with arbitrary FIXMessage.

        Args:
            msg: arbitrary FIXMessage
        """
        if self.schema:
            self.schema.validate(msg)

        raw_msg = self.conn_accept._codec.encode(
            msg,
            self.conn_accept._session,
            raw_seq_num=FTag.MsgSeqNum in msg,
        ).encode()

        # Pretend the message was transfered to initiator
        decoded_msg, _, _ = self.conn_init._codec.decode(raw_msg, silent=False)

        if self.schema:
            self.schema.validate(decoded_msg)

        self.acceptor_sent.append(decoded_msg)

        await self.conn_init._process_message(decoded_msg, raw_msg)

        return decoded_msg

    def _next_order_id(self) -> int:
        self._order_id += 1
        return self._order_id

    def _next_exec_id(self) -> int:
        self._exec_id += 1
        return self._exec_id

    def order_register_single(self, o: FIXNewOrderSingle):
        """Registers FIXNewOrderSingle."""
        self.registered_orders[o.clord_id] = o
        return True

    def fix_cxl_request(self, o: FIXNewOrderSingle) -> FIXMessage:
        """Generates FIXNewOrderSingle cancel request + optional validation.

        Returns:
            cancel FIXMessage
        """
        assert o.can_cancel()  # new assert 2023-09-23
        m = o.cancel_req()
        if self.schema:
            self.schema.validate(m)
        self.registered_orders[o.clord_id] = o
        return m

    def fix_rep_request(
        self,
        o: FIXNewOrderSingle,
        price: float = nan,
        qty: float = nan,
    ) -> FIXMessage:
        """Generates FIXNewOrderSingle replace request + optional validating.

        Args:
            o: FIXNewOrderSingle
            price: new price, nan - to skip
            qty: new quantity, nan - to skip

        Returns:
            replace FIXMessage
        """
        assert o.can_replace()
        m = o.replace_req(price, qty)
        if self.schema:
            self.schema.validate(m)
        self.registered_orders[o.clord_id] = o
        return m

    def fix_cxlrep_reject_msg(
        self,
        cxl_req: FIXMessage,
        ord_status: FOrdStatus,
    ) -> FIXMessage:
        """Generates synthetic ORDERCANCELREJECT message.

        Args:
            cxl_req: cancel / replace request FIXMessage (from initiator)
            ord_status: new order status generated by this reply

        Returns:
            ORDERCANCELREJECT FIXMessage
        """
        clord_id = cxl_req[FTag.ClOrdID]
        orig_clord_id = cxl_req[FTag.OrigClOrdID]

        m = FIXMessage(FMsg.ORDERCANCELREJECT)
        m[37] = 0
        m[11] = clord_id
        m[41] = orig_clord_id
        m[39] = ord_status

        assert cxl_req.msg_type in [
            FMsg.ORDERCANCELREQUEST,
            FMsg.ORDERCANCELREPLACEREQUEST,
        ]

        if cxl_req.msg_type == FMsg.ORDERCANCELREQUEST:
            m[FTag.CxlRejResponseTo] = "1"
        elif cxl_req.msg_type == FMsg.ORDERCANCELREPLACEREQUEST:
            m[FTag.CxlRejResponseTo] = "2"

        if self.schema:
            self.schema.validate(m)

        return m

    def fix_exec_report_msg(
        self,
        order: FIXNewOrderSingle,
        clord_id: str,
        exec_type: FExecType,
        ord_status: FOrdStatus,
        cum_qty: float = nan,
        leaves_qty: float = nan,
        last_qty: float = nan,
        price: float = nan,
        order_qty: float = nan,
        orig_clord_id: str = None,
        avg_price: float = 0.0,
    ) -> FIXMessage:
        """Generates synthetic EXECUTIONREPORT.

        Args:
            order: FIXNewOrderSingle to report
            clord_id: report ClOrdID tag
            exec_type: report ExecType
            ord_status: report FOrdStatus
            cum_qty: filled qty
            leaves_qty: remaining qty
            last_qty: last trade qty
            price: (ONLY IN REPLACE) price of replaced order
            order_qty: (ONLY IN REPLACE) qty of replaced order
            orig_clord_id: report OrigClOrdID
            avg_price: average fill price

        Returns:
            EXECUTIONREPORT FIXMessage
        """
        assert order.clord_id in self.registered_orders, "Unregistered order!"

        m = FIXMessage(FMsg.EXECUTIONREPORT)
        assert clord_id
        m[FTag.ClOrdID] = clord_id

        if order.order_id is None:
            order_id = self._next_order_id()
        else:
            order_id = order.order_id

        m[FTag.OrderID] = order_id
        m[FTag.ExecID] = self._next_exec_id()

        if orig_clord_id:
            m[FTag.OrigClOrdID] = orig_clord_id
        m[FTag.ExecType] = exec_type
        m[FTag.OrdStatus] = ord_status
        m[FTag.Side] = order.side

        if isnan(order_qty):
            order_qty = order.qty
        else:
            assert (
                exec_type == FExecType.REPLACED
            ), "Only applicable to exec_type=5 (replace)"
            assert order_qty > 0

        if isnan(cum_qty):
            cum_qty = order.cum_qty
        else:
            assert cum_qty <= order.qty
            assert cum_qty >= 0
        m[FTag.CumQty] = cum_qty

        if isnan(leaves_qty):
            leaves_qty = order.leaves_qty
        else:
            assert leaves_qty >= 0
            assert leaves_qty <= order_qty

        m[FTag.LeavesQty] = leaves_qty
        assert (
            cum_qty + leaves_qty <= order_qty
        ), f"cum_qty[{cum_qty}] + leaves_qty[{leaves_qty}] <= order_qty[{order_qty}]"

        if not isnan(last_qty):
            assert not isnan(leaves_qty), "Must also set leaves_qty, when trade"
            assert not isnan(cum_qty), "Must also set cum_qty, when trade"
            assert (
                exec_type == FExecType.TRADE
            ), "Only applicable to exec_type=F (trade)"
            assert last_qty > 0
            m[FTag.LastQty] = last_qty
            assert (
                round(last_qty - (cum_qty - order.cum_qty), 3) == 0
            ), "Probably incorrect Trade qty"
        else:
            assert (
                exec_type != FExecType.TRADE
            ), "You must set last_qty when exec_type=F (trade)"

        if not isnan(price):
            assert (
                exec_type == FExecType.REPLACED
            ), "Only applicable to exec_type=5 (replace)"
        else:
            price = order.price

        order.set_instrument(m)

        order.set_price_qty(m, price, order_qty)
        m[FTag.AvgPx] = avg_price

        order.set_account(m)

        if (
            exec_type == FExecType.PENDING_CANCEL
            and ord_status == FOrdStatus.PENDING_CANCEL
        ):
            assert order.orig_clord_id
            assert order.clord_id
            assert order.clord_id != order.orig_clord_id
            assert order.cum_qty == cum_qty
            assert order.leaves_qty == leaves_qty

        if (
            ord_status == FOrdStatus.FILLED
            or ord_status == FOrdStatus.CANCELED
            or ord_status == FOrdStatus.REJECTED
            or ord_status == FOrdStatus.EXPIRED
        ):
            assert leaves_qty == 0, "New order report is finished, but LeavesQty != 0"

        if self.schema:
            self.schema.validate(m)
        return m

    def msg_logon(self, tags: dict | None = None) -> FIXMessage:
        """Generates message LOGON + schema validation.

        Args:
            tags: extra tags

        Returns:
            FIXMessage
        """
        msg = FIXMessage(FMsg.LOGON, tags)
        if FTag.EncryptMethod not in msg:
            msg.set(FTag.EncryptMethod, 0)
        if FTag.HeartBtInt not in msg:
            msg.set(FTag.HeartBtInt, 30)

        if self.schema:
            self.schema.validate(msg)

        return msg

    def msg_logout(self) -> FIXMessage:
        """Generates message LOGOUT + schema validation."""
        msg = FIXMessage(FMsg.LOGOUT)

        if self.schema:
            self.schema.validate(msg)

        return msg

    def msg_heartbeat(self, test_req_id=None) -> FIXMessage:
        """Generates message HEARTBEAT + schema validation.

        Args:
            test_req_id: if Heartbeat in reply to TESTREQUEST
        """
        msg = FIXMessage(FMsg.HEARTBEAT)
        if test_req_id is not None:
            msg[FTag.TestReqID] = test_req_id

        if self.schema:
            self.schema.validate(msg)

        return msg

    def msg_test_request(self, test_req_id) -> FIXMessage:
        """Generates message TESTREQUEST + schema validation.

        Args:
            test_req_id: unique test request id
        """
        msg = FIXMessage(FMsg.TESTREQUEST)
        msg[FTag.TestReqID] = test_req_id

        if self.schema:
            self.schema.validate(msg)

        return msg

    def msg_sequence_reset(
        self,
        msg_seq_num: int,
        new_seq_no: int,
        is_gap_fill: bool = False,
    ) -> FIXMessage:
        """Generates message SEQUENCERESET + schema validation.

        Args:
            msg_seq_num: FTag.MsgSeqNum value
            new_seq_no: FTag.NewSeqNo value
            is_gap_fill: FTag.GapFillFlag "Y"/"N"
        """
        msg = FIXMessage(FMsg.SEQUENCERESET)
        msg.set(FTag.MsgSeqNum, msg_seq_num)
        msg.set(FTag.GapFillFlag, "Y" if is_gap_fill else "N")
        msg.set(FTag.NewSeqNo, new_seq_no)
        if self.schema:
            self.schema.validate(msg)
        return msg

    def msg_resend_request(self, begin_seq_no, end_seq_no="0") -> FIXMessage:
        """Generates message RESENDREQUEST + schema validation.

        Args:
            begin_seq_no: seq no to start with
            end_seq_no: end seq no, "0" - means all
        """
        msg = FIXMessage(FMsg.RESENDREQUEST)
        msg.set(FTag.BeginSeqNo, str(begin_seq_no))
        msg.set(FTag.EndSeqNo, str(end_seq_no))
        if self.schema:
            self.schema.validate(msg)
        return msg



import asyncio
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from asyncfix import AsyncFIXDummyServer, ConnectionRole, ConnectionState, Journaler
from asyncfix.errors import FIXConnectionError
from asyncfix.protocol import FIXProtocol44


def test_init():
    log = MagicMock()
    j = Journaler()
    server = AsyncFIXDummyServer(
        FIXProtocol44(),
        "sender",
        "target",
        j,
        "local",
        777,
        heartbeat_period=12,
        logger=log,
    )

    assert isinstance(server.protocol, FIXProtocol44)
    assert server._session.sender_comp_id == "sender"
    assert server._session.target_comp_id == "target"
    assert server._journaler is j
    assert server._host == "local"
    assert server._port == 777
    assert server._heartbeat_period == 12
    assert server.log is log
    assert server.connection_role == ConnectionRole.ACCEPTOR


@pytest.mark.asyncio
async def test_connect():
    log = MagicMock()
    j = Journaler()
    server = AsyncFIXDummyServer(
        FIXProtocol44(),
        "sender",
        "target",
        j,
        "local",
        777,
        heartbeat_period=12,
        logger=log,
    )

    assert server.connection_state == ConnectionState.DISCONNECTED_NOCONN_TODAY
    with (
        patch("asyncio.start_server") as mock_start_server,
        patch("asyncfix.connection.AsyncFIXConnection.connect") as mock_base_connect,
        patch.object(server, "on_connect") as mock_on_connect,
    ):
        mock_srv = AsyncMock()
        mock_start_server.return_value = mock_srv

        await server.connect()

        assert mock_base_connect.await_count == 1
        assert not mock_on_connect.called
        assert server.connection_state == ConnectionState.DISCONNECTED_NOCONN_TODAY

        assert mock_start_server.await_count == 1
        assert len(mock_start_server.call_args[0]) == 3
        assert mock_start_server.call_args[0][1] == server._host
        assert mock_start_server.call_args[0][2] == server._port
        assert mock_start_server.call_args[1] == {}

        assert mock_srv.serve_forever.awaited

        reader = AsyncMock(asyncio.StreamReader)
        writer = AsyncMock(asyncio.StreamWriter)
        await server._handle_accept(reader, writer)
        assert not writer.close.called

        assert server.connection_state == ConnectionState.NETWORK_CONN_ESTABLISHED
        assert mock_on_connect.await_count == 1
        assert server._socket_writer == writer
        assert server._socket_reader == reader

        with pytest.raises(FIXConnectionError, match="Server already working"):
            await server.connect()

        await server._handle_accept(reader, writer)
        assert writer.close.call_count == 1
        assert writer.wait_closed.await_count == 1



"""FIXSession module."""
from asyncfix import FIXMessage, FMsg, FTag


class FIXSession:
    """Generic FIX Session container.

    Attributes:
        key: session DB id
        sender_comp_id: session sender
        target_comp_id: session target
        next_num_out: next expected seq num out
        next_num_in: next expected seq num in
    """

    def __init__(self, key, target_comp_id: str, sender_comp_id: str):
        """Initialize session.

        Args:
            key: session DB id
            target_comp_id: session target
            sender_comp_id: session sender
        """
        self.key = key
        """Session DB ID / key."""

        self.sender_comp_id = sender_comp_id
        self.target_comp_id = target_comp_id

        # Note: None is set intentionally, because session typically has to be
        #   loaded or created by Journaler class, therefore it sets session values
        #   internally
        self.next_num_out = None
        self.next_num_in = None

    def __hash__(self):
        """Hash by (target, sender)."""
        return hash((self.target_comp_id, self.sender_comp_id))

    def __eq__(self, o):
        """Equality by FIXSession or tuple(target, sender)."""
        if isinstance(o, FIXSession):
            return (
                self.target_comp_id == o.target_comp_id
                and self.sender_comp_id == o.sender_comp_id
            )
        elif isinstance(o, tuple):
            if len(o) != 2:
                return False
            return self.target_comp_id == o[0] and self.sender_comp_id == o[1]
        return False

    def __repr__(self):
        """Repr."""
        return (
            f"FIXSession(key={self.key},"
            f" target={self.target_comp_id} sender={self.sender_comp_id} InSN={self.next_num_in} OutSN={self.next_num_out})"  # noqa
        )

    def validate_comp_ids(self, target_comp_id: str, sender_comp_id: str) -> bool:
        """Ensure target_comp_id/sender_comp_id match.

        Args:
            target_comp_id: incoming target_comp_id
            sender_comp_id: incoming sender_comp_id

        Returns:
            bool
        """
        return (
            self.sender_comp_id == sender_comp_id
            and self.target_comp_id == target_comp_id
        )

    def allocate_next_num_out(self):
        """Increments next_num_out counter."""
        n = str(self.next_num_out)
        self.next_num_out += 1
        return n

    def set_next_num_in(self, msg: FIXMessage) -> int:
        """Sets next_num_in based on incoming FIXMessage.

        Args:
            msg: incoming message

        Returns:
            - if OK - current message seq_no
            - if ERROR - 0 or -1
        """
        if msg.msg_type == FMsg.SEQUENCERESET:
            if FTag.NewSeqNo not in msg:
                # Garbled message
                return 0
            seq_no = int(msg[FTag.NewSeqNo]) - 1

        else:
            if FTag.MsgSeqNum not in msg:
                # Garbled message
                return 0
            seq_no = int(msg[FTag.MsgSeqNum])

            if seq_no != self.next_num_in:
                # Gap detected
                return -1

        self.next_num_in = seq_no + 1

        return seq_no



import pytest

from asyncfix import FIXMessage, FMsg, FTag
from asyncfix.session import FIXSession


def test_init():
    s = FIXSession("1", "target", "sender")
    assert s.key == "1"
    assert s.target_comp_id == "target"
    assert s.sender_comp_id == "sender"

    assert s.next_num_in is None
    assert s.next_num_out is None


def test_hash():
    s = FIXSession("1", "target", "sender")
    assert hash(s) == hash((s.target_comp_id, s.sender_comp_id))


def test_equality():
    s1 = FIXSession("1", "target", "sender")
    s2 = FIXSession("2", "target", "sender")
    s3 = FIXSession("1", "target1", "sender")
    s4 = FIXSession("1", "target", "sender2")

    assert s1 == s2
    assert s1 != s3
    assert s1 != s4
    assert s1 == ("target", "sender")
    assert s1 != ("target",)
    assert s1 != ("target", "sender", "unexp")
    assert s1 != ("1target", "sender")
    assert s1 != ("target", "1sender")
    assert s1 != "target,sender"


def test_repr():
    s1 = FIXSession("1", "target", "sender")
    s1.next_num_out = 1
    s1.next_num_in = 1
    assert "FIXSession(key=1, target=target sender=sender InSN=1 OutSN=1)" == repr(s1)


def test_validate_compid():
    s1 = FIXSession("1", "target", "sender")

    assert s1.validate_comp_ids("target", "sender")
    assert not s1.validate_comp_ids("sender", "target")


def test_allocate_next_num_out():
    s1 = FIXSession("1", "target", "sender")
    s1.next_num_out = 1
    assert s1.allocate_next_num_out() == "1"
    assert s1.next_num_out == 2


def test_set_next_num_in():
    s1 = FIXSession("1", "target", "sender")

    s1.next_num_in = 10

    assert s1.set_next_num_in(FIXMessage(FMsg.SEQUENCERESET)) == 0
    assert s1.next_num_in == 10

    assert s1.set_next_num_in(FIXMessage(FMsg.SEQUENCERESET, {FTag.NewSeqNo: 4})) == 3
    assert s1.next_num_in == 4

    assert s1.set_next_num_in(FIXMessage(FMsg.LOGON)) == 0
    assert s1.next_num_in == 4

    assert s1.set_next_num_in(FIXMessage(FMsg.LOGON, {FTag.MsgSeqNum: 3})) == -1
    assert s1.next_num_in == 4

    assert s1.set_next_num_in(FIXMessage(FMsg.LOGON, {FTag.MsgSeqNum: 5})) == -1
    assert s1.next_num_in == 4

    assert s1.set_next_num_in(FIXMessage(FMsg.LOGON, {FTag.MsgSeqNum: 4})) == 4
    assert s1.next_num_in == 5



"""AsyncFIX protocol package."""
from .protocol_base import FIXProtocolBase
from .protocol_fix44 import FIXProtocol44
from .schema import FIXSchema
from .order_single import FIXNewOrderSingle
from .common import FExecType, FOrdType, FOrdSide, FOrdStatus



from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from asyncfix import AsyncFIXClient, ConnectionRole, ConnectionState, Journaler
from asyncfix.errors import FIXConnectionError
from asyncfix.protocol import FIXProtocol44


def test_init():
    log = MagicMock()
    j = Journaler()
    client = AsyncFIXClient(
        FIXProtocol44(),
        "sender",
        "target",
        j,
        "local",
        777,
        heartbeat_period=12,
        logger=log,
    )

    assert isinstance(client.protocol, FIXProtocol44)
    assert client._session.sender_comp_id == "sender"
    assert client._session.target_comp_id == "target"
    assert client._journaler is j
    assert client._host == "local"
    assert client._port == 777
    assert client._heartbeat_period == 12
    assert client.log is log
    assert client.connection_role == ConnectionRole.INITIATOR


@pytest.mark.asyncio
async def test_connect():
    log = MagicMock()
    j = Journaler()
    client = AsyncFIXClient(
        FIXProtocol44(),
        "sender",
        "target",
        j,
        "local",
        777,
        heartbeat_period=12,
        logger=log,
    )

    assert client.connection_state == ConnectionState.DISCONNECTED_NOCONN_TODAY
    with (
        patch("asyncio.open_connection") as mock_open_connection,
        patch("asyncfix.connection.AsyncFIXConnection.connect") as mock_base_connect,
        patch.object(client, "on_connect") as mock_on_connect,
    ):
        mock_open_connection.return_value = ("r", "w")

        await client.connect()
        assert mock_base_connect.await_count == 1
        assert mock_open_connection.await_count == 1
        assert mock_open_connection.call_args[0] == ("local", 777)
        assert mock_open_connection.call_args[1] == {}
        assert client.connection_state == ConnectionState.NETWORK_CONN_ESTABLISHED
        assert client._socket_reader == "r"
        assert client._socket_writer == "w"

        assert mock_on_connect.awaited
        assert mock_on_connect.call_args[0] == ()
        assert mock_on_connect.call_args[1] == {}

        with pytest.raises(FIXConnectionError, match="Socket already connected"):
            await client.connect()


@pytest.mark.asyncio
async def test_connect_exception():
    log = MagicMock()
    j = Journaler()
    client = AsyncFIXClient(
        FIXProtocol44(),
        "sender",
        "target",
        j,
        "local",
        777,
        heartbeat_period=12,
        logger=log,
    )

    assert client.connection_state == ConnectionState.DISCONNECTED_NOCONN_TODAY
    with (
        patch("asyncio.open_connection") as mock_open_connection,
        patch("asyncfix.connection.AsyncFIXConnection.connect") as mock_base_connect,
        patch.object(client, "on_connect") as mock_on_connect,
    ):
        mock_open_connection.side_effect = ValueError

        await client.connect()
        assert mock_open_connection.await_count == 1
        assert client.connection_state == ConnectionState.DISCONNECTED_BROKEN_CONN
        assert client._socket_reader is None
        assert client._socket_writer is None

        assert mock_on_connect.await_count == 0



"""Generic SQLite Journaler."""
import sqlite3

from asyncfix.errors import DuplicateSeqNoError, FIXMessageError
from asyncfix.message import MessageDirection
from asyncfix.session import FIXSession


class Journaler(object):
    """Tracks FIX message history."""

    def __init__(self, filename=None):
        """Initialize SQLite Journaler.

        Args:
            filename: path to file, or None - to make in-memory journal
        """
        if filename is None:
            self.conn = sqlite3.connect(":memory:")
        else:
            self.conn = sqlite3.connect(filename)

        self.cursor = self.conn.cursor()
        self.cursor.execute(
            "CREATE TABLE IF NOT EXISTS message("
            "seqNo INTEGER NOT NULL,"
            "session INTEGER NOT NULL,"
            "direction INTEGER NOT NULL,"
            "msg TEXT,"
            "PRIMARY KEY (seqNo, session, direction))"
        )

        self.cursor.execute(
            "CREATE TABLE IF NOT EXISTS session("
            "sessionId INTEGER PRIMARY KEY AUTOINCREMENT,"
            "targetCompId TEXT NOT NULL,"
            "senderCompId TEXT NOT NULL,"
            "outboundSeqNo INTEGER DEFAULT 0,"
            "inboundSeqNo INTEGER DEFAULT 0,"
            "UNIQUE (targetCompId, senderCompId))"
        )

    def sessions(self) -> dict[tuple[str, str], FIXSession]:
        """Loads all available sessions from journal."""
        sessions = {}
        self.cursor.execute(
            "SELECT sessionId, targetCompId, senderCompId, outboundSeqNo, inboundSeqNo"
            " FROM session"
        )
        for sessionInfo in self.cursor:
            session = FIXSession(sessionInfo[0], sessionInfo[1], sessionInfo[2])
            session.next_num_out = sessionInfo[3]
            session.next_num_in = sessionInfo[4] + 1
            sessions[(session.target_comp_id, session.sender_comp_id)] = session

        return sessions

    def create_or_load(self, target_comp_id, sender_comp_id) -> FIXSession:
        """Creates or loads new session with unique target_comp_id/sender_comp_id.

        Args:
            target_comp_id: session targetCompId
            sender_comp_id: session senderCompId

        Returns:
            FIXSession
        """
        try:
            self.cursor.execute(
                "INSERT INTO session(targetCompId, senderCompId) VALUES(?, ?)",
                (target_comp_id, sender_comp_id),
            )
            session_id = self.cursor.lastrowid
            self.conn.commit()
            session = FIXSession(session_id, target_comp_id, sender_comp_id)
            session.next_num_out = 1
            session.next_num_in = 1
        except sqlite3.IntegrityError:
            self.cursor.execute(
                "SELECT sessionId, targetCompId, senderCompId, outboundSeqNo, inboundSeqNo"  # noqa
                " FROM session WHERE targetCompId = ? AND senderCompId = ?",
                (target_comp_id, sender_comp_id),
            )
            session_data = next(self.cursor)
            session = FIXSession(session_data[0], session_data[1], session_data[2])
            session.next_num_out = session_data[3] + 1
            session.next_num_in = session_data[4] + 1
            # print(f"Journaler: Loaded session: {session}")
        return session

    @staticmethod
    def find_seq_no(msg: bytes) -> int:
        """Finds 34=<seqno> in serialized message.

        Args:
            msg: encoded fix message

        Returns:
            seq no value

        Raises:
            FIXMessageError: if not found or malformed message
        """
        try:
            i_start = msg.index(b"\x0134=")
            i_end = msg.index(b"\x01", i_start + 1)
            return int(msg[i_start + 4 : i_end])
        except Exception:
            raise FIXMessageError(f"tag 34 is not found or invalid, in message: {msg}")

    def set_seq_num(
        self,
        session: FIXSession,
        next_num_out: int | None = None,
        next_num_in: int | None = None,
    ):
        """Sets journal and session seq num.

        Args:
            session: target session
            next_num_out: new expected num out (optional)
            next_num_in: new expected num in (optional)
        """
        if next_num_out is not None:
            assert next_num_out > 0
            session.next_num_out = next_num_out
        else:
            next_num_out = session.next_num_out

        if next_num_in is not None:
            assert next_num_in > 0
            session.next_num_in = next_num_in
        else:
            next_num_in = session.next_num_in

        self.cursor.execute(
            "UPDATE session SET inboundSeqNo=?, outboundSeqNo=?  WHERE sessionId = ?",
            (next_num_in - 1, next_num_out - 1, session.key),
        )

        self.cursor.execute(
            "DELETE FROM message WHERE session = ? AND seqNo >= ? AND direction = ?",
            (session.key, next_num_in, MessageDirection.INBOUND.value),
        )
        self.cursor.execute(
            "DELETE FROM message WHERE session = ? AND seqNo >= ? AND direction = ?",
            (session.key, next_num_out, MessageDirection.OUTBOUND.value),
        )

    def persist_msg(
        self,
        msg: bytes,
        session: FIXSession,
        direction: MessageDirection,
    ):
        """Commits encoded fix message into DB.

        Args:
            msg: encoded fix message
            session: target session
            direction: message direction

        Raises:
            DuplicateSeqNoError: (critical) when DB already have such message seq_no
        """
        assert isinstance(msg, bytes), "expected encoded message"
        seq_no = self.find_seq_no(msg)
        try:
            self.cursor.execute(
                "INSERT INTO message VALUES(?, ?, ?, ?)",
                (seq_no, session.key, direction.value, msg),
            )
            if direction == MessageDirection.OUTBOUND:
                self.cursor.execute(
                    "UPDATE session SET outboundSeqNo=? WHERE sessionId = ?",
                    (seq_no, session.key),
                )
            elif direction == MessageDirection.INBOUND:
                self.cursor.execute(
                    "UPDATE session SET inboundSeqNo=? WHERE sessionId = ?",
                    (seq_no, session.key),
                )

            self.conn.commit()
        except sqlite3.IntegrityError as e:
            raise DuplicateSeqNoError("%s is a duplicate, error %s" % (seq_no, repr(e)))

    def recover_msg(
        self,
        session: FIXSession,
        direction: MessageDirection,
        seq_no: int,
    ) -> bytes:
        """Loads specific message from DB by seq no.

        Args:
            session: target session
            direction: message direction
            seq_no: target seq_no

        Returns:
            encoded message (bytes)
        """
        msgs = self.recover_messages(session, direction, seq_no, seq_no)
        if msgs:
            return msgs[0]
        else:
            return None

    def recover_messages(
        self,
        session: FIXSession,
        direction: MessageDirection,
        start_seq_no: int | str,
        end_seq_no: int | str,
    ) -> list[bytes]:
        """Loads messages with seq no range from DB.

        Args:
            session: target session
            direction: message direction
            start_seq_no: seq no from
            end_seq_no: seq no to

        Returns:
            list of encoded FIX messages
        """
        self.cursor.execute(
            "SELECT msg FROM message WHERE session = ? AND direction = ? AND seqNo >= ?"
            " AND seqNo <= ? ORDER BY seqNo",
            (session.key, direction.value, start_seq_no, end_seq_no),
        )
        msgs = []
        for msg in self.cursor:
            assert isinstance(msg[0], bytes)
            msgs.append(msg[0])
        return msgs

    def get_all_msgs(
        self,
        sessions: list[FIXSession] | None = None,
        direction: MessageDirection | None = None,
    ):
        """Get all messages from the Journaler DB.

        Args:
            sessions: session filter (optional)
            direction: direction filter (optional)

        Returns:
            list of tuples [(seq_no, enc_msg, direction, session_key), ...]
        """
        sql = "SELECT seqNo, msg, direction, session FROM message"
        clauses = []
        args = []
        if sessions is not None and len(sessions) != 0:
            skeys = []
            for s in sessions:
                key = s.key if isinstance(s, FIXSession) else s
                skeys.append(key)

            clauses.append("session in (" + ",".join("?" * len(sessions)) + ")")
            args.extend(skeys)

        if direction is not None:
            clauses.append("direction = ?")
            args.append(direction.value)

        if clauses:
            sql = sql + " WHERE " + " AND ".join(clauses)

        sql = sql + " ORDER BY rowid"

        self.cursor.execute(sql, tuple(args))
        msgs = []
        for msg in self.cursor:
            msgs.append((msg[0], msg[1], msg[2], msg[3]))

        return msgs



"""Dummy FIX server module."""
import asyncio
import logging

from asyncfix.connection import AsyncFIXConnection, ConnectionRole, ConnectionState
from asyncfix.errors import FIXConnectionError
from asyncfix.journaler import Journaler
from asyncfix.protocol import FIXProtocolBase


class AsyncFIXDummyServer(AsyncFIXConnection):
    """Simple server which supports only single connection (just for testing)."""

    def __init__(
        self,
        protocol: FIXProtocolBase,
        sender_comp_id: str,
        target_comp_id: str,
        journaler: Journaler,
        host: str,
        port: int,
        heartbeat_period: int = 30,
        logger: logging.Logger | None = None,
    ):
        """Initialization.

        Args:
            protocol: FIX protocol used in codec
            sender_comp_id: server sender_comp_id tag
            target_comp_id: server target_comp_id tag
            journaler: message journaler
            host: fix host to listen
            port: fix port to bind
            heartbeat_period: heartbeat_period in seconds
            logger: custom logger instance
        """
        super().__init__(
            protocol=protocol,
            sender_comp_id=sender_comp_id,
            target_comp_id=target_comp_id,
            journaler=journaler,
            host=host,
            port=port,
            heartbeat_period=heartbeat_period,
            logger=logger,
        )
        self._connection_role = ConnectionRole.ACCEPTOR

    async def connect(self):
        """Starts the server and infinitely runs it.

        Raises:
            FIXConnectionError: when already connected
        """
        if self._socket_reader:
            raise FIXConnectionError("Server already working")

        # this must be called in order to launch socket_reader/heartbeat tasks
        await super().connect()

        server = await asyncio.start_server(self._handle_accept, self._host, self._port)
        async with server:
            await server.serve_forever()

    async def _handle_accept(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ):
        """Executed when new client connection established."""
        if self._socket_writer:
            self.log.info("Multiple connections are not allowed for dummy server")

            writer.close()
            await writer.wait_closed()

        self._socket_reader = reader
        self._socket_writer = writer
        self._connection_state = ConnectionState.NETWORK_CONN_ESTABLISHED
        addr = writer.get_extra_info("peername")
        self.log.info("Connection from %s" % repr(addr))

        await self.on_connect()



"""Base FIX protocol."""
from asyncfix import FMsg, FTag


class FIXProtocolBase(object):
    """Generic FIX protocol."""

    beginstring: str = "FIXProtocolBase"
    repeating_groups: dict[str, list[str]] = {}
    session_message_types: set = set()
    fixtags: FTag = FTag
    msgtype: FMsg = FMsg



import os
import time

import pytest

from asyncfix.connection import MessageDirection
from asyncfix.errors import DuplicateSeqNoError, FIXMessageError
from asyncfix.journaler import Journaler
from asyncfix.message import FIXContainer, FIXMessage
from asyncfix.session import FIXSession


def test_create_or_load():
    j = Journaler()

    s1 = j.create_or_load("test_target", "test_sender")

    assert isinstance(s1, FIXSession)

    assert s1.sender_comp_id == "test_sender"
    assert s1.target_comp_id == "test_target"
    assert s1.next_num_out == 1
    assert s1.next_num_in == 1

    assert s1.key == 1

    s2 = j.create_or_load("test_target", "test_sender")

    assert isinstance(s1, FIXSession)

    assert s2.sender_comp_id == "test_sender"
    assert s2.target_comp_id == "test_target"

    assert s2.next_num_out == 1
    assert s2.next_num_in == 1
    assert s2.key == 1

    assert s1 is not s2


def test_find_seqno():
    enc_msg = b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0134=073\x0152=20230919-07:13:26.808\x0144=123.45\x0138\x0155=VOD.L\x0110=100\x01"  # noqa
    assert Journaler.find_seq_no(enc_msg) == 73

    with pytest.raises(FIXMessageError):
        enc_msg = b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0134=\x0152=20230919-07:13:26.808\x0144=123.45\x0138\x0155=VOD.L\x0110=100\x01"  # noqa
        Journaler.find_seq_no(enc_msg)

    with pytest.raises(FIXMessageError):
        enc_msg = b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0152=20230919-07:13:26.808\x0144=123.45\x0138\x0155=VOD.L\x0110=100\x01"  # noqa
        Journaler.find_seq_no(enc_msg)

    with pytest.raises(FIXMessageError):
        enc_msg = b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0134=asdf\x0152=20230919-07:13:26.808\x0144=123.45\x0138\x0155=VOD.L\x0110=100\x01"  # noqa
        Journaler.find_seq_no(enc_msg)


def test_persist():
    enc_msg = b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0134=073\x0152=20230919-07:13:26.808\x0144=123.45\x0138\x0155=VOD.L\x0110=100\x01"  # noqa

    assert Journaler.find_seq_no(enc_msg) == 73

    j = Journaler()
    s1 = j.create_or_load("test_target", "test_sender")
    assert s1.next_num_out == 1
    assert s1.next_num_in == 1

    # persist and reload rewrites session seq num

    j.persist_msg(enc_msg, s1, MessageDirection.INBOUND)

    s2 = j.create_or_load("test_target", "test_sender")

    assert s2.next_num_in == 74
    assert s2.next_num_out == 1

    j.persist_msg(enc_msg, s1, MessageDirection.OUTBOUND)

    s2 = j.create_or_load("test_target", "test_sender")

    assert s2.next_num_in == 74
    assert s2.next_num_out == 74

    with pytest.raises(DuplicateSeqNoError, match="73 is a duplicate"):
        j.persist_msg(enc_msg, s1, MessageDirection.OUTBOUND)


def test_sessions():
    enc_msg = b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0134=073\x0152=20230919-07:13:26.808\x0144=123.45\x0138\x0155=VOD.L\x0110=100\x01"  # noqa

    assert Journaler.find_seq_no(enc_msg) == 73

    j = Journaler()

    s1 = j.create_or_load("test_target", "test_sender")

    s2 = j.create_or_load("test_target2", "test_sender2")

    # persist and reload rewrites session seq num

    j.persist_msg(enc_msg, s1, MessageDirection.INBOUND)

    sessions = j.sessions()

    assert isinstance(sessions, dict)

    assert ("test_target", "test_sender") in sessions
    assert s1 in sessions

    s_new = j.create_or_load("test_target", "test_sender")

    assert s_new in sessions


def test_persist_recover():
    enc_msg = b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0134=073\x0152=20230919-07:13:26.808\x0144=123.45\x0138\x0155=VOD.L\x0110=100\x01"  # noqa

    assert Journaler.find_seq_no(enc_msg) == 73

    j = Journaler()

    s1 = j.create_or_load("test_target", "test_sender")

    # persist and reload rewrites session seq num

    j.persist_msg(enc_msg, s1, MessageDirection.INBOUND)

    j.persist_msg(enc_msg, s1, MessageDirection.OUTBOUND)

    s2 = j.create_or_load("test_target", "test_sender")
    assert s2.next_num_in == 74
    assert s2.next_num_out == 74

    m = j.recover_msg(s2, MessageDirection.INBOUND, 73)
    assert m == enc_msg

    m2 = j.recover_msg(s2, MessageDirection.INBOUND, 173)

    assert m2 is None


def test_get_all_msg():
    enc_msg_in = b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0134=073\x0152=20230919-07:13:26.808\x0144=123.45\x0138\x0155=VOD.L\x0110=100\x01"  # noqa

    enc_msg_out = b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0134=078\x0152=20230919-07:13:26.808\x0144=123.45\x0138\x0155=VOD.L\x0110=100\x01"  # noqa

    j = Journaler()

    s1 = j.create_or_load("test_target", "test_sender")

    s2 = j.create_or_load("test_target2", "test_sender2")

    # persist and reload rewrites session seq num

    j.persist_msg(enc_msg_in, s1, MessageDirection.INBOUND)

    j.persist_msg(enc_msg_out, s1, MessageDirection.OUTBOUND)

    all = j.get_all_msgs()
    assert len(all) == 2

    assert all[0] == (73, enc_msg_in, MessageDirection.INBOUND.value, 1)

    assert all[1] == (78, enc_msg_out, MessageDirection.OUTBOUND.value, 1)

    all = j.get_all_msgs(sessions=[s1])
    assert len(all) == 2

    assert all[0] == (73, enc_msg_in, MessageDirection.INBOUND.value, 1)

    assert all[1] == (78, enc_msg_out, MessageDirection.OUTBOUND.value, 1)

    all = j.get_all_msgs(sessions=[s2])
    assert len(all) == 0

    all = j.get_all_msgs(sessions=[s1], direction=MessageDirection.INBOUND)
    assert len(all) == 1

    assert all[0] == (73, enc_msg_in, MessageDirection.INBOUND.value, 1)

    all = j.get_all_msgs(sessions=[s1], direction=MessageDirection.OUTBOUND)
    assert len(all) == 1

    assert all[0] == (78, enc_msg_out, MessageDirection.OUTBOUND.value, 1)


def test_persist_tofile():
    enc_msg = b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0134=073\x0152=20230919-07:13:26.808\x0144=123.45\x0138\x0155=VOD.L\x0110=100\x01"  # noqa

    j = Journaler("test.store")
    assert os.path.exists("test.store")
    os.unlink("test.store")


def test_seq_set():
    enc_messages = [
        b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0134=073\x0152=20230919-07:13:26.808\x0144=123.45\x0138\x0155=VOD.L\x0110=100\x01",  # noqa
        b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0134=074\x0152=20230919-07:13:26.808\x0144=123.45\x0138\x0155=VOD.L\x0110=100\x01",  # noqa
        b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0134=075\x0152=20230919-07:13:26.808\x0144=123.45\x0138\x0155=VOD.L\x0110=100\x01",  # noqa
        b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0134=076\x0152=20230919-07:13:26.808\x0144=123.45\x0138\x0155=VOD.L\x0110=100\x01",  # noqa
    ]
    assert Journaler.find_seq_no(enc_messages[0]) == 73

    j = Journaler("test.store")

    s1 = j.create_or_load("test_target", "test_sender")

    assert s1.next_num_in == 1
    assert s1.next_num_out == 1

    # persist and reload rewrites session seq num
    j.persist_msg(enc_messages[0], s1, MessageDirection.INBOUND)
    j.persist_msg(enc_messages[1], s1, MessageDirection.INBOUND)
    j.persist_msg(enc_messages[2], s1, MessageDirection.OUTBOUND)
    j.persist_msg(enc_messages[3], s1, MessageDirection.OUTBOUND)

    with pytest.raises(DuplicateSeqNoError, match="UNIQUE constraint failed"):
        j.persist_msg(enc_messages[0], s1, MessageDirection.INBOUND)

    s1 = j.create_or_load("test_target", "test_sender")
    assert s1.next_num_in == 75
    assert s1.next_num_out == 77

    j.set_seq_num(s1, next_num_out=76, next_num_in=74)
    assert s1.next_num_in == 74
    assert s1.next_num_out == 76

    s2 = j.create_or_load("test_target", "test_sender")
    assert s2.next_num_in == 74
    assert s2.next_num_out == 76

    with pytest.raises(DuplicateSeqNoError, match="UNIQUE constraint failed"):
        j.persist_msg(enc_messages[0], s1, MessageDirection.INBOUND)
    with pytest.raises(DuplicateSeqNoError, match="UNIQUE constraint failed"):
        j.persist_msg(enc_messages[2], s1, MessageDirection.OUTBOUND)

    j.persist_msg(enc_messages[1], s1, MessageDirection.INBOUND)
    j.persist_msg(enc_messages[3], s1, MessageDirection.OUTBOUND)

    j.set_seq_num(s2)
    assert s2.next_num_in == 74
    assert s2.next_num_out == 76

    with pytest.raises(DuplicateSeqNoError, match="UNIQUE constraint failed"):
        j.persist_msg(enc_messages[0], s1, MessageDirection.INBOUND)
    with pytest.raises(DuplicateSeqNoError, match="UNIQUE constraint failed"):
        j.persist_msg(enc_messages[2], s1, MessageDirection.OUTBOUND)

    j.persist_msg(enc_messages[1], s1, MessageDirection.INBOUND)
    j.persist_msg(enc_messages[3], s1, MessageDirection.OUTBOUND)

    os.unlink("test.store")


def test_perf():
    msg_tpl = b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0134=076\x0152=20230919-07:13:26.808\x0144=123.45\x0138\x0155=VOD.L\x0110=100\x01"  # noqa

    j = Journaler("test.store")

    s1 = j.create_or_load("test_target", "test_sender")

    # persist and reload rewrites session seq num
    t_begin = time.time()
    n_msgs = 10
    for i in range(1, n_msgs):
        _m = msg_tpl.replace(b"34=076", f"34={i}".encode())
        j.persist_msg(_m, s1, MessageDirection.INBOUND)
    t_end = time.time()
    os.unlink("test.store")
    # assert False, (t_end-t_begin) / n_msgs



"""Abstract connection module."""
import asyncio
import logging
import sys
import time
from enum import Enum, IntEnum

from asyncfix import FMsg, FTag
from asyncfix.codec import Codec
from asyncfix.errors import FIXConnectionError
from asyncfix.journaler import Journaler
from asyncfix.message import FIXMessage, MessageDirection
from asyncfix.protocol import FIXProtocolBase
from asyncfix.session import FIXSession


# fmt: off
class ConnectionState(IntEnum):
    """Connection status enum."""
    UNKNOWN = 0

    DISCONNECTED_NOCONN_TODAY = 1    # both
    """Currently disconnected, have not attempted to establish a connection today."""

    DISCONNECTED_WCONN_TODAY = 2     # both
    """Currently disconnected, have attempted to establish a connection today."""

    DISCONNECTED_BROKEN_CONN = 3    # both
    """While connected, detect a broken network connection (e.g. TCP socket closed)."""

    AWAITING_CONNECTION = 4          # acceptor
    """Session acceptor Logon awaiting network connection from counterparty."""

    INITIATE_CONNECTION = 5          # initiator
    """Session initiator Logon establishing network connection with counterparty."""

    NETWORK_CONN_ESTABLISHED = 6     # both
    """Network connection established between both parties."""

    LOGON_INITIAL_SENT = 7           # initiator
    """Session initiator Logon send Logon(35=A) message."""

    LOGON_INITIAL_RECV = 8           # acceptor
    """Session acceptor Logon receive counterparty’s Logon(35=A) message."""

    LOGON_RESPONSE = 9               # acceptor
    """Session acceptor Logon respond to peer with Logon message to handshake."""

    RESENDREQ_HANDLING = 10            # both
    """Receive and respond to counterparty’s ResendRequest(35=2) sending requested
       messages and/or SequenceReset(35=4) gap fill messages for the range of
       MsgSeqNum(34) requested.
    """

    RECV_SEQNUM_TOO_HIGH = 11        # both
    """Receive too high of MsgSeqNum(34) from counterparty, queue message,
       and send ResendRequest(35=2).
    """

    RESENDREQ_AWAITING = 12          # both
    """Process requested MsgSeqNum(34) with PossDupFlag(43)=Y resent messages and/or
    SequenceReset(35=4) gap fill messages from counterparty.
    """

    NO_MSG_IN_INTERVAL = 13          # both
    """No inbound messages (non-garbled) received in (HeartBtInt+ reasonable time."""

    AWAIT_PROC_TEST_REQ = 14         # both
    """Process inbound messages.

    Reset heartbeat interval-related timer when ANY
       inbound message (non-garbled) is received.
    """

    RECEIVED_LOGOUT = 15             # both
    """Receive Logout(35=5) message from counterparty initiating logout/disconnect."""

    INITIATE_LOGOUT = 16             # both
    """Identify condition or reason to gracefully disconnect (e.g. end of “day”,
    no response after multiple TestRequest(35=1) messages, too low MsgSeqNum(34)).
    """

    ACTIVE = 17                      # both
    """Network connection established, Logon(35=A) message exchange completed."""

    WAITING_FOR_LOGON = 18           # initiator
    """Session initiator waiting for session acceptor to send back a Logon(35=A)."""
# fmt: on


class ConnectionRole(Enum):
    """Role of the connection INITIATOR / ACCEPTOR."""

    UNKNOWN = 0
    INITIATOR = 1
    ACCEPTOR = 2


class AsyncFIXConnection:
    """AsyncFIX bidirectional connection.

    Attributes:
        connection_state: Current connection_state
        connection_role: Current connection_role ACCEPTOR | INITIATOR
        log: logger
    """

    def __init__(
        self,
        protocol: FIXProtocolBase,
        sender_comp_id: str,
        target_comp_id: str,
        journaler: Journaler,
        host: str,
        port: int,
        heartbeat_period: int = 30,
        logger: logging.Logger | None = None,
    ):  # noqa: D403
        """AsyncFIX bidirectional connection.

        Args:
            protocol: FIX protocol
            sender_comp_id: initiator SenderCompID
            target_comp_id: acceptor TargetCompID
            journaler: fix messages journaling engine
            host: endpoint host
            port: endpoint port
            heartbeat_period: heartbeat interval in seconds
            logger: logger instance (by default logging.getLogger())
            start_tasks: True - starts socket/heartbeat asyncio tasks, False - no tasks
                        (this is useful in debugging / testing)
        """
        if not logger:
            self.log: logging.Logger = logging.getLogger()
        else:
            self.log: logging.Logger = logger

        # Private attributes
        self._connection_state: ConnectionState = (
            ConnectionState.DISCONNECTED_NOCONN_TODAY
        )
        self._connection_role: ConnectionRole = ConnectionRole.UNKNOWN
        self._codec = Codec(protocol)
        self._journaler: Journaler = journaler
        self._session: FIXSession = journaler.create_or_load(
            target_comp_id, sender_comp_id
        )
        self._connection_was_active = False
        self._msg_buffer = b""
        self._heartbeat_period = heartbeat_period
        self._message_last_time = 0.0
        self._max_seq_num_resend = 0
        self._test_req_id = None
        self._socket_reader: asyncio.StreamReader = None
        self._socket_writer: asyncio.StreamWriter = None
        self._host = host
        self._port = int(port)
        self._aio_task_socket_read = None
        self._aio_task_heartbeat = None

    @property
    def connection_state(self) -> ConnectionState:
        """Current connection state."""
        return self._connection_state

    @property
    def connection_role(self) -> ConnectionRole:
        """Current connection role."""
        return self._connection_role

    @property
    def heartbeat_period(self) -> int:
        """Current connection heartbeat period in seconds."""
        return self._heartbeat_period

    @property
    def protocol(self) -> FIXProtocolBase:
        """Underlying FIXProtocolBase of a connection."""
        return self._codec.protocol

    async def connect(self):
        """Transport initialization method."""
        if not self._aio_task_socket_read:
            self._aio_task_socket_read = asyncio.create_task(self.socket_read_task())
        if not self._aio_task_heartbeat:
            self._aio_task_heartbeat = asyncio.create_task(self.heartbeat_timer_task())

    async def disconnect(
        self,
        disconn_state: ConnectionState,
        logout_message: str = None,
    ):
        """Disconnect session and closes the socket.

        Args:
            disconn_state: connection state after disconnection
            logout_message: if not None, sends Logout() message to peer with
                            (58=logout_message)
        """
        if self._connection_state > ConnectionState.DISCONNECTED_BROKEN_CONN:
            assert disconn_state <= ConnectionState.DISCONNECTED_BROKEN_CONN
            self._test_req_id = None
            self._message_last_time = 0.0
            self._max_seq_num_resend = 0

            if logout_message is not None:
                msg = FIXMessage(FMsg.LOGOUT)
                if logout_message:
                    # Only add message if logout_message != ""
                    msg[FTag.Text] = logout_message
                await self.send_msg(msg)

            self.log.info(f"Client disconnected, with state: {repr(disconn_state)}")
            if self._socket_writer:
                self._socket_writer.close()
                await self._socket_writer.wait_closed()
            self._socket_writer = None
            self._socket_reader = None
            await self._state_set(disconn_state)
            await self.on_disconnect()

    async def send_msg(self, msg: FIXMessage):
        """Sends message to the peer.

        Args:
            msg: fix message

        Raises:
            FIXConnectionError: raised if connection state does not allow sending
        """
        if self._connection_state < ConnectionState.NETWORK_CONN_ESTABLISHED:
            raise FIXConnectionError(
                "Connection must be established before sending any "
                f"FIX message, got state: {repr(self._connection_state)}"
            )
        elif self._connection_state == ConnectionState.NETWORK_CONN_ESTABLISHED:
            # INITIATOR mode, connection was just established
            if msg.msg_type != FMsg.LOGON and msg.msg_type != FMsg.LOGOUT:
                raise FIXConnectionError(
                    "You must send first Logon(35=A)/Logout() message immediately after"
                    f" connection, got {repr(msg)}"
                )
            await self._state_set(ConnectionState.LOGON_INITIAL_SENT)
            self._connection_role = ConnectionRole.INITIATOR
        else:
            if self._connection_role == ConnectionRole.INITIATOR:
                if (
                    self._connection_state == ConnectionState.LOGON_INITIAL_SENT
                    and msg.msg_type != FMsg.LOGOUT
                ):
                    raise FIXConnectionError(
                        "Initiator is waiting for Logon() response, you must not send"
                        " any additional messages before acceptor responce."
                    )

        if msg.msg_type == FMsg.TESTREQUEST and self._test_req_id is None:
            raise FIXConnectionError(
                "You must rend TestRequest() message via self.send_test_req() method in"
                " order to get valid response handling"
            )

        encoded_msg = self._codec.encode(msg, self._session).encode("utf-8")

        msg_raw = encoded_msg.replace(b"\x01", b"|")
        self.log.debug(
            f"[{self._connection_role.name}]:send_msg ({self._connection_state.name})"
            f" {repr(msg.msg_type)}\n\t {msg_raw.decode()}\n"
        )

        self._socket_writer.write(encoded_msg)
        await self._socket_writer.drain()

        self._journaler.persist_msg(
            encoded_msg, self._session, MessageDirection.OUTBOUND
        )

    async def send_test_req(self):
        """Sends TestRequest(35=1) and sets TestReqID for expected response from peer.

        Raises:
            FIXConnectionError: if another TestRequest() is pending
        """
        if self._test_req_id is not None:
            raise FIXConnectionError("Another test request already pending")

        self._test_req_id = int(time.time())
        test_msg = FIXMessage(FMsg.TESTREQUEST, {FTag.TestReqID: self._test_req_id})
        await self.send_msg(test_msg)

    async def socket_read_task(self):
        """Main socket reader task (decode raw messages and calls _process_message)."""
        last_connect = time.time()

        while True:
            try:
                if not self._socket_reader:
                    # Socket was not connected, just wait
                    await asyncio.sleep(1)
                    if self.connection_role == ConnectionRole.INITIATOR:
                        if time.time() - last_connect > self.heartbeat_period * 1.5:
                            self.log.debug("Trying reconnect")
                            last_connect = time.time()
                            await self.connect()
                    continue

                msg = await self._socket_reader.read(4096)
                if not msg:
                    raise ConnectionError

                self._msg_buffer = self._msg_buffer + msg
                while True:
                    if (
                        self._connection_state
                        <= ConnectionState.DISCONNECTED_BROKEN_CONN
                    ):
                        break

                    (decoded_msg, parsed_length, raw_msg) = self._codec.decode(
                        self._msg_buffer
                    )
                    # self.log.debug(decoded_msg)

                    if parsed_length > 0:
                        self._msg_buffer = self._msg_buffer[parsed_length:]

                    if decoded_msg is None:
                        break

                    await self._process_message(decoded_msg, raw_msg)
            except asyncio.CancelledError:
                return
            except ConnectionError as why:
                self.log.debug(
                    "socket_read_task: connection has been closed %s" % (why,)
                )
                await self.disconnect(ConnectionState.DISCONNECTED_BROKEN_CONN)

                # Reset reconnect time
                last_connect = time.time()
                continue
            except Exception:
                self.log.exception("socket_read_task: exception")
                # raise

    async def heartbeat_timer_task(self):
        """Heartbeat watcher task."""
        while True:
            try:
                if not self._socket_writer or not self._socket_reader:
                    # Socket was not connected, just wait
                    await asyncio.sleep(1)
                    continue

                tm = time.time()

                if self._connection_state == ConnectionState.ACTIVE:
                    if tm - self._message_last_time > self._heartbeat_period - 1:
                        if not self._test_req_id:
                            await self.send_test_req()
                        self._message_last_time = tm

                if (
                    self._message_last_time
                    and tm - self._message_last_time > self._heartbeat_period * 2
                ):
                    # Dead socket probably
                    self.log.debug("heartbeat_timer_task: message last time timeout")
                    await self.disconnect(ConnectionState.DISCONNECTED_BROKEN_CONN)

                if (
                    self._test_req_id
                    and tm - self._test_req_id > self._heartbeat_period * 2
                ):
                    # No sensible reply on TestRequest
                    self.log.debug("heartbeat_timer_task: test request timeout")
                    await self.disconnect(ConnectionState.DISCONNECTED_BROKEN_CONN)

                await asyncio.sleep(1.0)
            except asyncio.CancelledError:
                return
            except Exception:
                self.log.exception("heartbeat_timer() error")

    async def reset_seq_num(self):
        """Resets session and journal seq nums to 1."""
        self.log.info("Resetting connection sequence and journal")
        self._journaler.set_seq_num(self._session, next_num_in=1, next_num_out=1)
        assert self._session.next_num_in == 1
        assert self._session.next_num_out == 1

    ####################################################
    #
    #  User Application methods
    #
    ####################################################

    async def on_message(self, msg: FIXMessage):
        """(AppEvent) Business message was received.

        Typically excludes session messages

        Args:
            msg: generic incoming FIXMessage
        """
        raise NotImplementedError("on_message() must be implemented in app class")

    async def on_connect(self):
        """(AppEvent) Underlying socket connected."""
        raise NotImplementedError("on_connect() must be implemented in app class")
        pass

    async def on_disconnect(self):
        """(AppEvent) Underlying socket disconnected."""
        pass

    async def on_logon(self, is_healthy: bool):
        """(AppEvent) Logon(35=A) received from peer.

        Args:
            is_healthy: True - if connection_state is ACTIVE
        """
        pass

    async def on_logout(self, msg: FIXMessage):
        """(AppEvent) Logout(35=5) received from peer.

        Args:
            msg: Logout(35=5) FIXMessage
        """
        pass

    async def on_state_change(self, connection_state: ConnectionState):
        """(AppEvent) On ConnectionState change.

        Args:
            connection_state: new connection state
        """
        pass

    async def should_replay(self, historical_replay_msg: FIXMessage) -> bool:
        """(AppLevel) Checks if historical_replay_msg from Journaler should be replayed.

        Args:
            historical_replay_msg: message from Journaler log

        Returns: True - replay, False - msg skipped (replaced by SequenceReset(35=4))
        """
        return True

    ####################################################
    #
    #  Private methods
    #
    ####################################################

    async def _state_set(self, connection_state: ConnectionState):
        """Sets internal connection state.

        Args:
            connection_state: new connection state
        """
        self.log.debug(
            f"[{self._connection_role.name}] NewState: {connection_state.name}"
        )
        self._connection_state = connection_state
        if connection_state == ConnectionState.ACTIVE:
            self._connection_was_active = True
        await self.on_state_change(connection_state)

    def _validate_integrity(self, msg: FIXMessage) -> bool:
        """Validates incoming message critical integrity.

        Args:
            msg: incoming message

        Returns:
            None - if no error
            True - if critical error (no Logout() message has to be sent)
            "err msg" - Logout(58="err msg") should be sent
        """
        if msg[FTag.BeginString] != self.protocol.beginstring:
            return (
                "Protocol BeginString(8) mismatch, expected"
                f" {self.protocol.beginstring}, got {msg[FTag.BeginString]}"
            )
        if FTag.SenderCompID not in msg or FTag.TargetCompID not in msg:
            # this will drop connection without a message
            return True

        if not self._session.validate_comp_ids(
            msg[FTag.SenderCompID], msg[FTag.TargetCompID]
        ):
            # Sender/Target are reversed here
            return "TargetCompID / SenderCompID mismatch"

        # TODO: validate SendingTime ~~ within 2x heartbeat_period
        if FTag.MsgSeqNum not in msg:
            return "MsgSeqNum(34) tag is missing"

        msg_seq_num = int(msg[FTag.MsgSeqNum])
        if msg_seq_num < self._session.next_num_in:
            _is_err = True
            if msg.msg_type == FMsg.SEQUENCERESET:
                _is_err = False
            if self._connection_state == ConnectionState.RESENDREQ_AWAITING:
                _is_err = False

            if _is_err:
                return (
                    f"MsgSeqNum is too low, expected {self._session.next_num_in}, got"
                    f" {msg_seq_num}"
                )

        # All good
        return None

    async def _process_logon(self, logon_msg: FIXMessage):
        """Processes Logon(35=A) message."""
        assert logon_msg.msg_type == FMsg.LOGON
        assert (
            self._connection_role == ConnectionRole.ACCEPTOR
            or self._connection_role == ConnectionRole.INITIATOR
        )

        msg_seq_num = int(logon_msg[FTag.MsgSeqNum])

        if self._connection_role == ConnectionRole.ACCEPTOR:
            assert self._connection_state == ConnectionState.LOGON_INITIAL_RECV
            if msg_seq_num >= self._session.next_num_in:
                msg_logon = FIXMessage(FMsg.LOGON)
                msg_logon.set(FTag.EncryptMethod, logon_msg[FTag.EncryptMethod])
                msg_logon.set(FTag.HeartBtInt, logon_msg[FTag.HeartBtInt])
                await self.send_msg(msg_logon)

        if msg_seq_num == self._session.next_num_in:
            await self._state_set(ConnectionState.ACTIVE)
        else:
            await self._state_set(ConnectionState.RECV_SEQNUM_TOO_HIGH)

        await self.on_logon(self._connection_state == ConnectionState.ACTIVE)

    async def _check_seqnum_gaps(self, msg_seq_num: int) -> bool:
        """Validates incoming MsgSeqNum, sends ResendRequest(35=2) if gap.

        Args:
            msg_seq_num:

        Returns:
            True - no gap, MsgSeqNum is correct
            False - there is a gap, ResendRequest() sent
        """
        if msg_seq_num > self._session.next_num_in:
            if self._connection_state != ConnectionState.RESENDREQ_AWAITING:
                resend_req = FIXMessage(
                    FMsg.RESENDREQUEST,
                    {FTag.BeginSeqNo: self._session.next_num_in, FTag.EndSeqNo: "0"},
                )
                self._max_seq_num_resend = msg_seq_num
                await self.send_msg(resend_req)
                await self._state_set(ConnectionState.RESENDREQ_AWAITING)
            return False

        return True

    async def _process_logout(self, logout_msg: FIXMessage):
        """Processes incoming Logout(35=5) message.

        Args:
            logout_msg: Logout(35=5) FIXMessage
        """
        assert logout_msg.msg_type == FMsg.LOGOUT

        if self._connection_was_active:
            dstate = ConnectionState.DISCONNECTED_WCONN_TODAY
        else:
            dstate = ConnectionState.DISCONNECTED_BROKEN_CONN

        await self.on_logout(logout_msg)

        await self.disconnect(dstate)

    async def _process_resend(self, resend_msg: FIXMessage):
        """Handles ResendRequest(35=2) - fills message gaps.

        Args:
            resend_msg: ResendRequest(35=2) FIXMessage
        """
        if self._connection_state != ConnectionState.RESENDREQ_AWAITING:
            await self._state_set(ConnectionState.RESENDREQ_HANDLING)

        assert resend_msg.msg_type == FMsg.RESENDREQUEST
        assert self._connection_state in {
            ConnectionState.RESENDREQ_HANDLING,
            ConnectionState.RESENDREQ_AWAITING,
        }

        begin_seq_no = int(resend_msg[FTag.BeginSeqNo])
        end_seq_no = int(resend_msg[FTag.EndSeqNo])
        if end_seq_no == 0:
            end_seq_no = sys.maxsize
        self.log.info("Received resent request from %s to %s", begin_seq_no, end_seq_no)
        journal_replay_msgs = self._journaler.recover_messages(
            self._session, MessageDirection.OUTBOUND, begin_seq_no, end_seq_no
        )

        # Remember next_num_out
        current_next_num_out = self._session.next_num_out

        self._journaler.set_seq_num(self._session, next_num_out=begin_seq_no)
        gap_fill_begin = int(begin_seq_no)
        gap_fill_end = int(begin_seq_no)

        noreply_msgs = {
            FMsg.LOGON,
            FMsg.LOGOUT,
            FMsg.RESENDREQUEST,
            FMsg.HEARTBEAT,
            FMsg.TESTREQUEST,
            FMsg.SEQUENCERESET,
        }

        for enc_msg in journal_replay_msgs:
            replay_msg, _, _ = self._codec.decode(enc_msg, silent=False)
            msg_seq_num = int(replay_msg[FTag.MsgSeqNum])

            is_sess_msg = replay_msg[FTag.MsgType] in noreply_msgs
            if is_sess_msg or not await self.should_replay(replay_msg):
                gap_fill_end = msg_seq_num + 1
            else:
                if gap_fill_begin < gap_fill_end:
                    # we need to send a gap fill message
                    gap_fill_msg = FIXMessage(FMsg.SEQUENCERESET)
                    gap_fill_msg[FTag.GapFillFlag] = "Y"
                    gap_fill_msg[FTag.MsgSeqNum] = gap_fill_begin
                    gap_fill_msg[FTag.NewSeqNo] = str(gap_fill_end)
                    # breakpoint()
                    await self.send_msg(gap_fill_msg)

                # and then resent the replayMsg
                replay_msg[FTag.PossDupFlag] = "Y"
                replay_msg[FTag.OrigSendingTime] = replay_msg[FTag.SendingTime]
                del replay_msg[FTag.MsgType]
                del replay_msg[FTag.BeginString]
                del replay_msg[FTag.BodyLength]
                del replay_msg[FTag.SendingTime]
                del replay_msg[FTag.SenderCompID]
                del replay_msg[FTag.TargetCompID]
                del replay_msg[FTag.CheckSum]
                await self.send_msg(replay_msg)

                gap_fill_begin = msg_seq_num + 1

        if gap_fill_end < gap_fill_begin:
            self.log.warning(
                "Journal MsgSeqNum not reflecting last"
                f" next_num_out={current_next_num_out}, forcing reset."
            )

        assert gap_fill_end <= current_next_num_out, "Unexpected end for gap"

        # Remainder not available in some reason
        if gap_fill_begin < current_next_num_out:
            gap_fill_msg = FIXMessage(FMsg.SEQUENCERESET)
            gap_fill_msg[FTag.GapFillFlag] = "Y"
            gap_fill_msg[FTag.MsgSeqNum] = gap_fill_begin
            gap_fill_msg[FTag.NewSeqNo] = current_next_num_out
            await self.send_msg(gap_fill_msg)

        self._journaler.set_seq_num(self._session, next_num_out=current_next_num_out)

        if self._connection_state != ConnectionState.RESENDREQ_AWAITING:
            await self._state_set(ConnectionState.ACTIVE)

    async def _process_seqreset(self, seqreset_msg: FIXMessage):
        """Handles SequenceReset(35=4) message.

        Args:
            seqreset_msg: SequenceReset(35=4) FIXMessage
        """
        assert seqreset_msg.msg_type == FMsg.SEQUENCERESET

        if seqreset_msg.get(FTag.GapFillFlag, None) == "Y":
            if self._connection_state != ConnectionState.RESENDREQ_AWAITING:
                self.log.warning(
                    "Getting SEQUENCERESET(GapFillFlag=Y) while not filling gaps"
                )
        else:
            self.log.info(f"SequenceReset received from peer: {seqreset_msg}")

        # Cleanup journal of past messages if session was reset to avoid SQL dup errors
        #   Sometimes we might have outdated seq nums in journal
        self._journaler.set_seq_num(
            self._session, next_num_in=int(seqreset_msg[FTag.MsgSeqNum])
        )

        # Set journal at new NewSeqNo
        self._journaler.set_seq_num(
            self._session, next_num_in=int(seqreset_msg[FTag.NewSeqNo])
        )

    async def _finalize_message(self, msg: FIXMessage, raw_msg: bytes):
        """Final message processing (MsgSeqNum checks / journaling).

        Args:
            msg: incoming message
            raw_msg: encoded message for journal
        """
        msg_sec_no = self._session.set_next_num_in(msg)

        if msg_sec_no <= 0:
            self.log.warning(f"Trying to finalize invalid {msg=}")
            return

        if self._connection_state == ConnectionState.RESENDREQ_AWAITING:
            assert self._max_seq_num_resend > 0

            if msg_sec_no >= self._max_seq_num_resend:
                # All messages were transferred
                self._max_seq_num_resend = 0
                await self._state_set(ConnectionState.ACTIVE)

        self._message_last_time = time.time()

        self._journaler.persist_msg(raw_msg, self._session, MessageDirection.INBOUND)

    async def _process_testrequest(self, testreq_msg: FIXMessage):
        """Handles TestRequest(35=1).

        Replies with Heartbeat(35=0) message with TestReqID tag

        Args:
            testreq_msg: TestRequest FIXMessage
        """
        assert testreq_msg.msg_type == FMsg.TESTREQUEST
        hbt_msg = FIXMessage(
            FMsg.HEARTBEAT, {FTag.TestReqID: testreq_msg.get(FTag.TestReqID, 0)}
        )
        await self.send_msg(hbt_msg)

    async def _process_heartbeat(self, hbt_msg: FIXMessage):
        """Handles Heartbeat(35=0) message (possible response on TestRequest(35=1)).

        Args:
            hbt_msg: Heartbeat(35=0) FIXMessage
        """
        assert hbt_msg.msg_type == FMsg.HEARTBEAT

        if self._test_req_id is not None:
            if FTag.TestReqID not in hbt_msg:
                # Possibly interval Heartbeat, just skip
                return

            # Expecting test_req_id
            try:
                msg_test_id = int(hbt_msg.get(FTag.TestReqID, "0"))
            except Exception:
                msg_test_id = 0

            if self._test_req_id != msg_test_id:
                await self.disconnect(
                    ConnectionState.DISCONNECTED_BROKEN_CONN,
                    logout_message="Invalid TestRequest(TestReqID) received",
                )
            else:
                self._test_req_id = None

    async def _process_message(self, msg: FIXMessage, raw_msg: bytes):
        """Main message processing dispatcher.

        Args:
            msg: incoming decoded message
            raw_msg: incoming raw message (bytes)
        """
        self.log.debug(
            f"[{self._connection_role.name}]:process_message"
            f" ({self._connection_state.name}) {repr(msg.msg_type)}\n\t {msg}\n"
        )

        err_msg = self._validate_integrity(msg)
        if err_msg:
            # Some mandatory tags are missing or corrupt message
            await self.disconnect(
                ConnectionState.DISCONNECTED_BROKEN_CONN,
                logout_message=err_msg if isinstance(err_msg, str) else None,
            )
            return
        is_valid_msg_num = False
        try:
            assert self._connection_state >= ConnectionState.NETWORK_CONN_ESTABLISHED

            if self._connection_state == ConnectionState.NETWORK_CONN_ESTABLISHED:
                # Applicable only for acceptor
                if msg.msg_type != FMsg.LOGON:
                    # According to FIX session spec, first message must be Logon()
                    #  we must drop connection without Logout() if first message was
                    #  not logon
                    await self.disconnect(ConnectionState.DISCONNECTED_BROKEN_CONN)
                    return
                await self._state_set(ConnectionState.LOGON_INITIAL_RECV)
                self._connection_role = ConnectionRole.ACCEPTOR

            if msg.msg_type == FMsg.LOGON:
                await self._process_logon(msg)
            elif msg.msg_type == FMsg.SEQUENCERESET:
                await self._process_seqreset(msg)
            elif msg.msg_type == FMsg.LOGOUT:
                await self._process_logout(msg)

            if self._connection_state <= ConnectionState.DISCONNECTED_BROKEN_CONN:
                # Got logout probably
                return

            msg_seq_num = int(msg[FTag.MsgSeqNum])
            is_valid_msg_num = await self._check_seqnum_gaps(msg_seq_num)

            if msg.msg_type == FMsg.RESENDREQUEST:
                await self._process_resend(msg)
            elif msg.msg_type == FMsg.SEQUENCERESET:
                pass
            elif msg.msg_type == FMsg.LOGON:
                pass  # already processed
            elif msg.msg_type == FMsg.TESTREQUEST:
                await self._process_testrequest(msg)
            elif msg.msg_type == FMsg.HEARTBEAT:
                await self._process_heartbeat(msg)
            else:
                if is_valid_msg_num:
                    await self.on_message(msg)
                else:
                    self.log.debug(f"_process_message: skipped app msg: {msg}")

        except asyncio.CancelledError:
            raise
        except Exception:
            self.log.exception(
                f"[{self._connection_role.name}]:process_message"
                f" ({self._connection_state.name}) {repr(msg.msg_type)}\n\t {msg}\n"
            )
        finally:
            if is_valid_msg_num:
                await self._finalize_message(msg, raw_msg)
